import numpy as np
import torch

from ritm_annotation.model.metrics import _compute_iou

from .brs_losses import BRSMaskLoss


class BaseOptimizer:
    def __init__(
        self,
        optimizer_params,
        prob_thresh=0.49,
        reg_weight=1e-3,
        min_iou_diff=0.01,
        brs_loss=BRSMaskLoss(),
        with_flip=False,
        flip_average=False,
        **kwargs
    ):
        self.brs_loss = brs_loss
        self.optimizer_params = optimizer_params
        self.prob_thresh = prob_thresh
        self.reg_weight = reg_weight
        self.min_iou_diff = min_iou_diff
        self.with_flip = with_flip
        self.flip_average = flip_average

        self.best_prediction = None
        self._get_prediction_logits = None
        self._opt_shape = None
        self._best_loss = None
        self._click_masks = None
        self._last_mask = None
        self.device = None

    def init_click(
        self, get_prediction_logits, pos_mask, neg_mask, device, shape=None
    ):
        self.best_prediction = None
        self._get_prediction_logits = get_prediction_logits
        self._click_masks = (pos_mask, neg_mask)
        self._opt_shape = shape
        self._last_mask = None
        self.device = device

    def __call__(self, x):
        opt_params = torch.from_numpy(x).float().to(self.device)
        opt_params.requires_grad_(True)

        with torch.enable_grad():
            opt_vars, reg_loss = self.unpack_opt_params(opt_params)
            result_before_sigmoid = self._get_prediction_logits(*opt_vars)
            result = torch.sigmoid(result_before_sigmoid)

            pos_mask, neg_mask = self._click_masks
            if self.with_flip and self.flip_average:
                result, result_flipped = torch.chunk(result, 2, dim=0)
                result = 0.5 * (result + torch.flip(result_flipped, dims=[3]))
                pos_mask, neg_mask = (
                    pos_mask[: result.shape[0]],
                    neg_mask[: result.shape[0]],
                )

            loss, f_max_pos, f_max_neg = self.brs_loss(
                result, pos_mask, neg_mask
            )
            loss = loss + reg_loss

        f_val = loss.detach().cpu().numpy()
        if self.best_prediction is None or f_val < self._best_loss:
            self.best_prediction = result_before_sigmoid.detach()
            self._best_loss = f_val

        if f_max_pos < (1 - self.prob_thresh) and f_max_neg < self.prob_thresh:
            return [f_val, np.zeros_like(x)]

        current_mask = result > self.prob_thresh
        if self._last_mask is not None and self.min_iou_diff > 0:
            diff_iou = _compute_iou(current_mask, self._last_mask)
            if len(diff_iou) > 0 and diff_iou.mean() > 1 - self.min_iou_diff:
                return [f_val, np.zeros_like(x)]
        self._last_mask = current_mask

        loss.backward()
        f_grad = opt_params.grad.cpu().numpy().ravel().astype(np.float)

        return [f_val, f_grad]

    def unpack_opt_params(self, opt_params):
        raise NotImplementedError


class InputOptimizer(BaseOptimizer):
    def unpack_opt_params(self, opt_params):
        opt_params = opt_params.view(self._opt_shape)
        if self.with_flip:
            opt_params_flipped = torch.flip(opt_params, dims=[3])
            opt_params = torch.cat([opt_params, opt_params_flipped], dim=0)
        reg_loss = self.reg_weight * torch.sum(opt_params**2)

        return (opt_params,), reg_loss


class ScaleBiasOptimizer(BaseOptimizer):
    def __init__(self, *args, scale_act=None, reg_bias_weight=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale_act = scale_act
        self.reg_bias_weight = reg_bias_weight

    def unpack_opt_params(self, opt_params):
        scale, bias = torch.chunk(opt_params, 2, dim=0)
        reg_loss = self.reg_weight * (
            torch.sum(scale**2) + self.reg_bias_weight * torch.sum(bias**2)
        )

        if self.scale_act == "tanh":
            scale = torch.tanh(scale)
        elif self.scale_act == "sin":
            scale = torch.sin(scale)

        return (1 + scale, bias), reg_loss



from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class GrabCutDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        images_dir_name="data_GT",
        masks_dir_name="boundary_GT",
        **kwargs
    ):
        super(GrabCutDataset, self).__init__(**kwargs)

        self.dataset_path = Path(dataset_path)
        self._images_path = self.dataset_path / images_dir_name
        self._insts_path = self.dataset_path / masks_dir_name

        self.dataset_samples = [
            x.name for x in sorted(self._images_path.glob("*.*"))
        ]
        self._masks_paths = {x.stem: x for x in self._insts_path.glob("*.*")}

    def get_sample(self, index) -> DSample:
        image_name = self.dataset_samples[index]
        image_path = str(self._images_path / image_name)
        mask_path = str(self._masks_paths[image_name.split(".")[0]])

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = cv2.imread(mask_path)[:, :, 0].astype(np.int32)
        instances_mask[instances_mask == 128] = -1
        instances_mask[instances_mask > 128] = 1

        return DSample(
            image,
            instances_mask,
            objects_ids=[1],
            ignore_ids=[-1],
            sample_id=index,
        )



import numpy as np
import torch
import torch.nn as nn

from ritm_annotation.model.modifiers import LRMult
from ritm_annotation.model.ops import BatchImageNormalize, DistMaps, ScaleLayer


class ISModel(nn.Module):
    def __init__(
        self,
        use_rgb_conv=True,
        with_aux_output=False,
        norm_radius=260,
        use_disks=False,
        cpu_dist_maps=False,
        clicks_groups=None,
        with_prev_mask=False,
        use_leaky_relu=False,
        binary_prev_mask=False,
        conv_extend=False,
        norm_layer=nn.BatchNorm2d,
        norm_mean_std=([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
    ):
        super().__init__()
        self.with_aux_output = with_aux_output
        self.clicks_groups = clicks_groups
        self.with_prev_mask = with_prev_mask
        self.binary_prev_mask = binary_prev_mask
        self.normalization = BatchImageNormalize(
            norm_mean_std[0], norm_mean_std[1]
        )

        self.coord_feature_ch = 2
        if clicks_groups is not None:
            self.coord_feature_ch *= len(clicks_groups)

        if self.with_prev_mask:
            self.coord_feature_ch += 1

        if use_rgb_conv:
            rgb_conv_layers = [
                nn.Conv2d(
                    in_channels=3 + self.coord_feature_ch,
                    out_channels=6 + self.coord_feature_ch,
                    kernel_size=1,
                ),
                norm_layer(6 + self.coord_feature_ch),
                nn.LeakyReLU(negative_slope=0.2)
                if use_leaky_relu
                else nn.ReLU(inplace=True),
                nn.Conv2d(
                    in_channels=6 + self.coord_feature_ch,
                    out_channels=3,
                    kernel_size=1,
                ),
            ]
            self.rgb_conv = nn.Sequential(*rgb_conv_layers)
        elif conv_extend:
            self.rgb_conv = None
            self.maps_transform = nn.Conv2d(
                in_channels=self.coord_feature_ch,
                out_channels=64,
                kernel_size=3,
                stride=2,
                padding=1,
            )
            self.maps_transform.apply(LRMult(0.1))
        else:
            self.rgb_conv = None
            mt_layers = [
                nn.Conv2d(
                    in_channels=self.coord_feature_ch,
                    out_channels=16,
                    kernel_size=1,
                ),
                nn.LeakyReLU(negative_slope=0.2)
                if use_leaky_relu
                else nn.ReLU(inplace=True),
                nn.Conv2d(
                    in_channels=16,
                    out_channels=64,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                ),
                ScaleLayer(init_value=0.05, lr_mult=1),
            ]
            self.maps_transform = nn.Sequential(*mt_layers)

        if self.clicks_groups is not None:
            self.dist_maps = nn.ModuleList()
            for click_radius in self.clicks_groups:
                self.dist_maps.append(
                    DistMaps(
                        norm_radius=click_radius,
                        spatial_scale=1.0,
                        cpu_mode=cpu_dist_maps,
                        use_disks=use_disks,
                    )
                )
        else:
            self.dist_maps = DistMaps(
                norm_radius=norm_radius,
                spatial_scale=1.0,
                cpu_mode=cpu_dist_maps,
                use_disks=use_disks,
            )

    def forward(self, image, points):
        image, prev_mask = self.prepare_input(image)
        coord_features = self.get_coord_features(image, prev_mask, points)

        if self.rgb_conv is not None:
            x = self.rgb_conv(torch.cat((image, coord_features), dim=1))
            outputs = self.backbone_forward(x)
        else:
            coord_features = self.maps_transform(coord_features)
            outputs = self.backbone_forward(image, coord_features)

        outputs["instances"] = nn.functional.interpolate(
            outputs["instances"],
            size=image.size()[2:],
            mode="bilinear",
            align_corners=True,
        )
        if self.with_aux_output:
            outputs["instances_aux"] = nn.functional.interpolate(
                outputs["instances_aux"],
                size=image.size()[2:],
                mode="bilinear",
                align_corners=True,
            )

        return outputs

    def prepare_input(self, image):
        prev_mask = None
        if self.with_prev_mask:
            prev_mask = image[:, 3:, :, :]
            image = image[:, :3, :, :]
            if self.binary_prev_mask:
                prev_mask = (prev_mask > 0.5).float()

        image = self.normalization(image)
        return image, prev_mask

    def backbone_forward(self, image, coord_features=None):
        raise NotImplementedError

    def get_coord_features(self, image, prev_mask, points):
        if self.clicks_groups is not None:
            points_groups = split_points_by_order(
                points,
                groups=(2,) + (1,) * (len(self.clicks_groups) - 2) + (-1,),
            )
            coord_features = [
                dist_map(image, pg)
                for dist_map, pg in zip(self.dist_maps, points_groups)
            ]
            coord_features = torch.cat(coord_features, dim=1)
        else:
            coord_features = self.dist_maps(image, points)

        if prev_mask is not None:
            coord_features = torch.cat((prev_mask, coord_features), dim=1)

        return coord_features


def split_points_by_order(tpoints: torch.Tensor, groups):
    points = tpoints.cpu().numpy()
    num_groups = len(groups)
    bs = points.shape[0]
    num_points = points.shape[1] // 2

    groups = [x if x > 0 else num_points for x in groups]
    group_points = [
        np.full((bs, 2 * x, 3), -1, dtype=np.float32) for x in groups
    ]

    last_point_indx_group = np.zeros((bs, num_groups, 2), dtype=np.int)  # type: ignore  # noqa: E501
    for group_indx, group_size in enumerate(groups):
        last_point_indx_group[:, group_indx, 1] = group_size

    for bindx in range(bs):
        for pindx in range(2 * num_points):
            point = points[bindx, pindx, :]
            group_id = int(point[2])
            if group_id < 0:
                continue

            is_negative = int(pindx >= num_points)
            if group_id >= num_groups or (
                group_id == 0 and is_negative
            ):  # disable negative first click
                group_id = num_groups - 1

            new_point_indx = last_point_indx_group[
                bindx, group_id, is_negative
            ]
            last_point_indx_group[bindx, group_id, is_negative] += 1

            group_points[group_id][bindx, new_point_indx, :] = point

    group_points = [
        torch.tensor(x, dtype=tpoints.dtype, device=tpoints.device)  # type: ignore  # noqa: E501
        for x in group_points
    ]

    return group_points



import json
import random
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class CocoDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        split="train",
        stuff_prob=0.0,
        dry_run=False,
        **kwargs,
    ):
        super(CocoDataset, self).__init__(**kwargs)
        self.split = split
        self.dataset_path = Path(dataset_path)
        self.stuff_prob = stuff_prob
        self.dry_run = dry_run

        self.load_samples()

    def load_samples(self):
        annotation_path = (
            self.dataset_path / "annotations" / f"panoptic_{self.split}.json"
        )
        self.labels_path = (
            self.dataset_path / "annotations" / f"panoptic_{self.split}"
        )
        self.images_path = self.dataset_path / self.split

        if not self.dry_run:
            with open(annotation_path, "r") as f:
                annotation = json.load(f)
        else:
            annotation = dict(annotations=[], categories=[])

        self.dataset_samples = annotation["annotations"]

        self._categories = annotation["categories"]
        self._stuff_labels = [
            x["id"] for x in self._categories if x["isthing"] == 0
        ]
        self._things_labels = [
            x["id"] for x in self._categories if x["isthing"] == 1
        ]
        self._things_labels_set = set(self._things_labels)
        self._stuff_labels_set = set(self._stuff_labels)

    def get_sample(self, index) -> DSample:
        dataset_sample = self.dataset_samples[index]

        image_path = self.images_path / self.get_image_name(
            dataset_sample["file_name"]
        )
        label_path = self.labels_path / dataset_sample["file_name"]

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        label = cv2.imread(str(label_path), cv2.IMREAD_UNCHANGED).astype(
            np.int32
        )
        label = (
            256 * 256 * label[:, :, 0] + 256 * label[:, :, 1] + label[:, :, 2]
        )

        instance_map = np.full_like(label, 0)
        things_ids = []
        stuff_ids = []

        for segment in dataset_sample["segments_info"]:
            class_id = segment["category_id"]
            obj_id = segment["id"]
            if class_id in self._things_labels_set:
                if segment["iscrowd"] == 1:
                    continue
                things_ids.append(obj_id)
            else:
                stuff_ids.append(obj_id)

            instance_map[label == obj_id] = obj_id

        if self.stuff_prob > 0 and random.random() < self.stuff_prob:
            instances_ids = things_ids + stuff_ids
        else:
            instances_ids = things_ids

            for stuff_id in stuff_ids:
                instance_map[instance_map == stuff_id] = 0

        return DSample(image, instance_map, objects_ids=instances_ids)

    @classmethod
    def get_image_name(cls, panoptic_name):
        return panoptic_name.replace(".png", ".jpg")



import torch

from ritm_annotation.model.losses import SigmoidBinaryCrossEntropyLoss


class BRSMaskLoss(torch.nn.Module):
    def __init__(self, eps=1e-5):
        super().__init__()
        self._eps = eps

    def forward(self, result, pos_mask, neg_mask):
        pos_diff = (1 - result) * pos_mask
        pos_target = torch.sum(pos_diff**2)
        pos_target = pos_target / (torch.sum(pos_mask) + self._eps)

        neg_diff = result * neg_mask
        neg_target = torch.sum(neg_diff**2)
        neg_target = neg_target / (torch.sum(neg_mask) + self._eps)

        loss = pos_target + neg_target

        with torch.no_grad():
            f_max_pos = torch.max(torch.abs(pos_diff)).item()
            f_max_neg = torch.max(torch.abs(neg_diff)).item()

        return loss, f_max_pos, f_max_neg


class OracleMaskLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.gt_mask = None
        self.loss = SigmoidBinaryCrossEntropyLoss(from_sigmoid=True)
        self.predictor = None
        self.history = []

    def set_gt_mask(self, gt_mask):
        self.gt_mask = gt_mask
        self.history = []

    def forward(self, result, pos_mask, neg_mask):
        gt_mask = self.gt_mask.to(result.device)
        if self.predictor.object_roi is not None:
            r1, r2, c1, c2 = self.predictor.object_roi[:4]
            gt_mask = gt_mask[:, :, r1 : r2 + 1, c1 : c2 + 1]
            gt_mask = torch.nn.functional.interpolate(
                gt_mask, result.size()[2:], mode="bilinear", align_corners=True
            )

        if result.shape[0] == 2:
            gt_mask_flipped = torch.flip(gt_mask, dims=[3])
            gt_mask = torch.cat([gt_mask, gt_mask_flipped], dim=0)

        loss = self.loss(result, gt_mask)
        self.history.append(loss.detach().cpu().numpy()[0])

        if (
            len(self.history) > 5
            and abs(self.history[-5] - self.history[-1]) < 1e-5
        ):
            return 0, 0, 0

        return loss, 1.0, 1.0



import os
import pickle as pkl
import random
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample
from ritm_annotation.utils.misc import get_labels_with_sizes


class ADE20kDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        split="train",
        stuff_prob=0.0,
        dry_run=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        assert split in {"train", "val"}

        self.dataset_path = Path(dataset_path)
        self.dataset_split = split
        self.dataset_split_folder = (
            "training" if split == "train" else "validation"
        )
        self.stuff_prob = stuff_prob

        anno_path = (
            self.dataset_path / f"{split}-annotations-object-segmentation.pkl"
        )
        if os.path.exists(anno_path):
            with anno_path.open("rb") as f:
                annotations = pkl.load(f)
        elif dry_run:
            annotations = {}
        else:
            raise RuntimeError(f"Can't find annotations at {anno_path}")
        self.annotations = annotations
        self.dataset_samples = list(annotations.keys())

    def get_sample(self, index) -> DSample:
        image_id = self.dataset_samples[index]
        sample_annos = self.annotations[image_id]

        image_path = str(
            self.dataset_path / sample_annos["folder"] / f"{image_id}.jpg"
        )
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # select random mask for an image
        layer = random.choice(sample_annos["layers"])
        mask_path = str(
            self.dataset_path / sample_annos["folder"] / layer["mask_name"]
        )
        instances_mask = cv2.imread(mask_path, cv2.IMREAD_UNCHANGED)[
            :, :, 0
        ]  # the B channel holds instances
        instances_mask = instances_mask.astype(np.int32)
        object_ids, _ = get_labels_with_sizes(instances_mask)

        if (self.stuff_prob <= 0) or (random.random() > self.stuff_prob):
            # remove stuff objects
            for i, object_id in enumerate(object_ids):
                if i in layer["stuff_instances"]:
                    instances_mask[instances_mask == object_id] = 0
            object_ids, _ = get_labels_with_sizes(instances_mask)

        return DSample(
            image, instances_mask, objects_ids=object_ids, sample_id=index
        )



"""
ritm_annotation base module.

This is the principal module of the ritm_annotation project.
here you put your main classes and objects.

Be creative! do whatever you want!

If you want to replace this with a Flask application run:

    $ make init

and then choose `flask` as template.
"""

# example constant variable
NAME = "ritm_annotation"



from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class ImagesDirDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        images_dir_name="images",
        masks_dir_name="masks",
        **kwargs
    ):
        super(ImagesDirDataset, self).__init__(**kwargs)

        self.dataset_path = Path(dataset_path)
        self._images_path = self.dataset_path / images_dir_name
        self._insts_path = self.dataset_path / masks_dir_name

        images_list = [x for x in sorted(self._images_path.glob("*.*"))]

        samples = {x.stem: {"image": x, "masks": []} for x in images_list}
        for mask_path in self._insts_path.glob("*.*"):
            mask_name = mask_path.stem
            if mask_name in samples:
                samples[mask_name]["masks"].append(mask_path)
                continue

            mask_name_split = mask_name.split("_")
            if mask_name_split[-1].isdigit():
                mask_name = "_".join(mask_name_split[:-1])
                assert mask_name in samples
                samples[mask_name]["masks"].append(mask_path)

        for x in samples.values():
            assert len(x["masks"]) > 0, x["image"]

        self.dataset_samples = [v for k, v in sorted(samples.items())]

    def get_sample(self, index) -> DSample:
        sample = self.dataset_samples[index]
        image_path = str(sample["image"])

        objects = []
        ignored_regions = []
        masks = []
        for indx, mask_path in enumerate(sample["masks"]):
            gt_mask = cv2.imread(str(mask_path))[:, :, 0].astype(np.int32)
            instances_mask = np.zeros_like(gt_mask)
            instances_mask[gt_mask == 128] = 2
            instances_mask[gt_mask > 128] = 1
            masks.append(instances_mask)
            objects.append((indx, 1))
            ignored_regions.append((indx, 2))

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        return DSample(
            image,
            np.stack(masks, axis=2),
            objects_ids=objects,
            ignore_ids=ignored_regions,
            sample_id=index,
        )



from tkinter import messagebox

import numpy as np
import torch

from ritm_annotation.inference import clicker
from ritm_annotation.inference.predictors import get_predictor
from ritm_annotation.utils.vis import draw_with_blend_and_clicks


class InteractiveController:
    def __init__(
        self,
        net,
        device,
        predictor_params,
        update_image_callback,
        prob_thresh=0.5,
    ):
        self.net = net
        self.prob_thresh = prob_thresh
        self.clicker = clicker.Clicker()
        self.states = []
        self.probs_history = []
        self.object_count = 0
        self._result_mask = None
        self._init_mask = None

        self.image = None
        self.predictor = None
        self.device = device
        self.update_image_callback = update_image_callback
        self.predictor_params = predictor_params
        self.reset_predictor()

    def set_image(self, image):
        self.image = image
        self._result_mask = np.zeros(image.shape[:2], dtype=np.uint16)
        self.object_count = 0
        self.reset_last_object(update_image=False)
        self.update_image_callback(reset_canvas=True)

    def set_mask(self, mask):
        if self.image.shape[:2] != mask.shape[:2]:
            messagebox.showwarning(
                "Warning",
                "A segmentation mask must have the same sizes as the current image!",  # noqa: E501
            )
            return

        if len(self.probs_history) > 0:
            self.reset_last_object()

        self._init_mask = mask.astype(np.float32)
        self.probs_history.append(
            (np.zeros_like(self._init_mask), self._init_mask)
        )
        self._init_mask = (
            torch.tensor(self._init_mask, device=self.device)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        self.clicker.click_indx_offset = 1

    def add_click(self, x, y, is_positive):
        self.states.append(
            {
                "clicker": self.clicker.get_state(),
                "predictor": self.predictor.get_states(),
            }
        )

        click = clicker.Click(is_positive=is_positive, coords=(y, x))
        self.clicker.add_click(click)
        pred = self.predictor.get_prediction(
            self.clicker, prev_mask=self._init_mask
        )
        if self._init_mask is not None and len(self.clicker) == 1:
            pred = self.predictor.get_prediction(
                self.clicker, prev_mask=self._init_mask
            )

        torch.cuda.empty_cache()

        if self.probs_history:
            self.probs_history.append((self.probs_history[-1][0], pred))
        else:
            self.probs_history.append((np.zeros_like(pred), pred))

        self.update_image_callback()

    def undo_click(self):
        if not self.states:
            return

        prev_state = self.states.pop()
        self.clicker.set_state(prev_state["clicker"])
        self.predictor.set_states(prev_state["predictor"])
        self.probs_history.pop()
        if not self.probs_history:
            self.reset_init_mask()
        self.update_image_callback()

    def partially_finish_object(self):
        object_prob = self.current_object_prob
        if object_prob is None:
            return

        self.probs_history.append((object_prob, np.zeros_like(object_prob)))
        self.states.append(self.states[-1])

        self.clicker.reset_clicks()
        self.reset_predictor()
        self.reset_init_mask()
        self.update_image_callback()

    def finish_object(self):
        if self.current_object_prob is None:
            return

        self._result_mask = self.result_mask
        self.object_count += 1
        self.reset_last_object()

    def reset_last_object(self, update_image=True):
        self.states = []
        self.probs_history = []
        self.clicker.reset_clicks()
        self.reset_predictor()
        self.reset_init_mask()
        if update_image:
            self.update_image_callback()

    def reset_predictor(self, predictor_params=None):
        if predictor_params is not None:
            self.predictor_params = predictor_params
        self.predictor = get_predictor(
            self.net, device=self.device, **self.predictor_params
        )
        if self.image is not None:
            self.predictor.set_input_image(self.image)

    def reset_init_mask(self):
        self._init_mask = None
        self.clicker.click_indx_offset = 0

    @property
    def current_object_prob(self):
        if self.probs_history:
            current_prob_total, current_prob_additive = self.probs_history[-1]
            return np.maximum(current_prob_total, current_prob_additive)
        else:
            return None

    @property
    def is_incomplete_mask(self):
        return len(self.probs_history) > 0

    @property
    def result_mask(self):
        result_mask = self._result_mask.copy()
        if self.probs_history:
            result_mask[self.current_object_prob > self.prob_thresh] = (
                self.object_count + 1
            )
        return result_mask

    def get_visualization(self, alpha_blend, click_radius):
        if self.image is None:
            return None

        results_mask_for_vis = self.result_mask
        vis = draw_with_blend_and_clicks(
            self.image,
            mask=results_mask_for_vis,
            alpha=alpha_blend,
            clicks_list=self.clicker.clicks_list,
            radius=click_radius,
        )
        if self.probs_history:
            total_mask = self.probs_history[-1][0] > self.prob_thresh
            results_mask_for_vis[np.logical_not(total_mask)] = 0
            vis = draw_with_blend_and_clicks(
                vis, mask=results_mask_for_vis, alpha=alpha_blend
            )

        return vis



import torch
import torch.nn.functional as F
from torchvision import transforms

from ritm_annotation.inference.transforms import (
    AddHorizontalFlip,
    LimitLongestSide,
    SigmoidForPred,
)


class BasePredictor(object):
    def __init__(
        self,
        model,
        device,
        net_clicks_limit=None,
        with_flip=False,
        zoom_in=None,
        max_size=None,
        **kwargs
    ):
        self.with_flip = with_flip
        self.net_clicks_limit = net_clicks_limit
        self.original_image = None
        self.device = device
        self.zoom_in = zoom_in
        self.prev_prediction = None
        self.model_indx = 0
        self.click_models = None
        self.net_state_dict = None

        if isinstance(model, tuple):
            self.net, self.click_models = model
        else:
            self.net = model

        self.to_tensor = transforms.ToTensor()

        self.transforms = [zoom_in] if zoom_in is not None else []
        if max_size is not None:
            self.transforms.append(LimitLongestSide(max_size=max_size))
        self.transforms.append(SigmoidForPred())
        if with_flip:
            self.transforms.append(AddHorizontalFlip())

    def set_input_image(self, image):
        image_nd = self.to_tensor(image)
        for transform in self.transforms:
            transform.reset()
        self.original_image = image_nd.to(self.device)
        if len(self.original_image.shape) == 3:
            self.original_image = self.original_image.unsqueeze(0)
        self.prev_prediction = torch.zeros_like(
            self.original_image[:, :1, :, :]
        )

    def get_prediction(self, clicker, prev_mask=None):
        clicks_list = clicker.get_clicks()

        if self.click_models is not None:
            model_indx = (
                min(
                    clicker.click_indx_offset + len(clicks_list),
                    len(self.click_models),
                )
                - 1
            )
            if model_indx != self.model_indx:
                self.model_indx = model_indx
                self.net = self.click_models[model_indx]

        input_image = self.original_image
        if prev_mask is None:
            prev_mask = self.prev_prediction
        if hasattr(self.net, "with_prev_mask") and self.net.with_prev_mask:
            input_image = torch.cat((input_image, prev_mask), dim=1)
        image_nd, clicks_lists, is_image_changed = self.apply_transforms(
            input_image, [clicks_list]
        )

        pred_logits = self._get_prediction(
            image_nd, clicks_lists, is_image_changed
        )
        prediction = F.interpolate(
            pred_logits,
            mode="bilinear",
            align_corners=True,
            size=image_nd.size()[2:],
        )

        for t in reversed(self.transforms):
            prediction = t.inv_transform(prediction)

        if (
            self.zoom_in is not None
            and self.zoom_in.check_possible_recalculation()
        ):
            return self.get_prediction(clicker)

        self.prev_prediction = prediction
        return prediction.cpu().numpy()[0, 0]

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        return self.net(image_nd, points_nd)["instances"]

    def _get_transform_states(self):
        return [x.get_state() for x in self.transforms]

    def _set_transform_states(self, states):
        assert len(states) == len(self.transforms)
        for state, transform in zip(states, self.transforms):
            transform.set_state(state)

    def apply_transforms(self, image_nd, clicks_lists):
        is_image_changed = False
        for t in self.transforms:
            image_nd, clicks_lists = t.transform(image_nd, clicks_lists)
            is_image_changed |= t.image_changed

        return image_nd, clicks_lists, is_image_changed

    def get_points_nd(self, clicks_lists):
        total_clicks = []
        num_pos_clicks = [
            sum(x.is_positive for x in clicks_list)
            for clicks_list in clicks_lists
        ]
        num_neg_clicks = [
            len(clicks_list) - num_pos
            for clicks_list, num_pos in zip(clicks_lists, num_pos_clicks)
        ]
        num_max_points = max(num_pos_clicks + num_neg_clicks)
        if self.net_clicks_limit is not None:
            num_max_points = min(self.net_clicks_limit, num_max_points)
        num_max_points = max(1, num_max_points)

        for clicks_list in clicks_lists:
            clicks_list = clicks_list[: self.net_clicks_limit]
            pos_clicks = [
                click.coords_and_indx
                for click in clicks_list
                if click.is_positive
            ]
            pos_clicks = pos_clicks + (num_max_points - len(pos_clicks)) * [
                (-1, -1, -1)
            ]

            neg_clicks = [
                click.coords_and_indx
                for click in clicks_list
                if not click.is_positive
            ]
            neg_clicks = neg_clicks + (num_max_points - len(neg_clicks)) * [
                (-1, -1, -1)
            ]
            total_clicks.append(pos_clicks + neg_clicks)

        return torch.tensor(total_clicks, device=self.device)

    def get_states(self):
        return {
            "transform_states": self._get_transform_states(),
            "prev_prediction": self.prev_prediction.clone(),
        }

    def set_states(self, states):
        self._set_transform_states(states["transform_states"])
        self.prev_prediction = states["prev_prediction"]



import torch.nn as nn

from ritm_annotation.model import ops


class ConvHead(nn.Module):
    def __init__(
        self,
        out_channels,
        in_channels=32,
        num_layers=1,
        kernel_size=3,
        padding=1,
        norm_layer=nn.BatchNorm2d,
    ):
        super(ConvHead, self).__init__()
        convhead = []

        for i in range(num_layers):
            convhead.extend(
                [
                    nn.Conv2d(
                        in_channels, in_channels, kernel_size, padding=padding
                    ),
                    nn.ReLU(),
                    norm_layer(in_channels)
                    if norm_layer is not None
                    else nn.Identity(),
                ]
            )
        convhead.append(nn.Conv2d(in_channels, out_channels, 1, padding=0))

        self.convhead = nn.Sequential(*convhead)

    def forward(self, *inputs):
        return self.convhead(inputs[0])


class SepConvHead(nn.Module):
    def __init__(
        self,
        num_outputs,
        in_channels,
        mid_channels,
        num_layers=1,
        kernel_size=3,
        padding=1,
        dropout_ratio=0.0,
        dropout_indx=0,
        norm_layer=nn.BatchNorm2d,
    ):
        super(SepConvHead, self).__init__()

        sepconvhead = []

        for i in range(num_layers):
            sepconvhead.append(
                SeparableConv2d(
                    in_channels=in_channels if i == 0 else mid_channels,
                    out_channels=mid_channels,
                    dw_kernel=kernel_size,
                    dw_padding=padding,
                    norm_layer=norm_layer,
                    activation="relu",
                )
            )
            if dropout_ratio > 0 and dropout_indx == i:
                sepconvhead.append(nn.Dropout(dropout_ratio))

        sepconvhead.append(
            nn.Conv2d(
                in_channels=mid_channels,
                out_channels=num_outputs,
                kernel_size=1,
                padding=0,
            )
        )

        self.layers = nn.Sequential(*sepconvhead)

    def forward(self, *inputs):
        x = inputs[0]

        return self.layers(x)


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        dw_kernel,
        dw_padding,
        dw_stride=1,
        activation=None,
        use_bias=False,
        norm_layer=None,
    ):
        super(SeparableConv2d, self).__init__()
        _activation = ops.select_activation_function(activation)
        self.body = nn.Sequential(
            nn.Conv2d(
                in_channels,
                in_channels,
                kernel_size=dw_kernel,
                stride=dw_stride,
                padding=dw_padding,
                bias=use_bias,
                groups=in_channels,
            ),
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=1,
                stride=1,
                bias=use_bias,
            ),
            norm_layer(out_channels)
            if norm_layer is not None
            else nn.Identity(),
            _activation(),
        )

    def forward(self, x):
        return self.body(x)



import os

import numpy as np
import torch
import torch._utils
import torch.nn as nn
import torch.nn.functional as F

from .ocr import SpatialGather_Module, SpatialOCR_Module
from .resnetv1b import BasicBlockV1b, BottleneckV1b

relu_inplace = True


class HighResolutionModule(nn.Module):
    def __init__(
        self,
        num_branches,
        blocks,
        num_blocks,
        num_inchannels,
        num_channels,
        fuse_method,
        multi_scale_output=True,
        norm_layer=nn.BatchNorm2d,
        align_corners=True,
    ):
        super(HighResolutionModule, self).__init__()
        self._check_branches(
            num_branches, num_blocks, num_inchannels, num_channels
        )

        self.num_inchannels = num_inchannels
        self.fuse_method = fuse_method
        self.num_branches = num_branches
        self.norm_layer = norm_layer
        self.align_corners = align_corners

        self.multi_scale_output = multi_scale_output

        self.branches = self._make_branches(
            num_branches, blocks, num_blocks, num_channels
        )
        self.fuse_layers = self._make_fuse_layers()
        self.relu = nn.ReLU(inplace=relu_inplace)

    def _check_branches(
        self, num_branches, num_blocks, num_inchannels, num_channels
    ):
        if num_branches != len(num_blocks):
            error_msg = "NUM_BRANCHES({}) <> NUM_BLOCKS({})".format(
                num_branches, len(num_blocks)
            )
            raise ValueError(error_msg)

        if num_branches != len(num_channels):
            error_msg = "NUM_BRANCHES({}) <> NUM_CHANNELS({})".format(
                num_branches, len(num_channels)
            )
            raise ValueError(error_msg)

        if num_branches != len(num_inchannels):
            error_msg = "NUM_BRANCHES({}) <> NUM_INCHANNELS({})".format(
                num_branches, len(num_inchannels)
            )
            raise ValueError(error_msg)

    def _make_one_branch(
        self, branch_index, block, num_blocks, num_channels, stride=1
    ):
        downsample = None
        if (
            stride != 1
            or self.num_inchannels[branch_index]
            != num_channels[branch_index] * block.expansion
        ):
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.num_inchannels[branch_index],
                    num_channels[branch_index] * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                self.norm_layer(num_channels[branch_index] * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.num_inchannels[branch_index],
                num_channels[branch_index],
                stride,
                downsample=downsample,
                norm_layer=self.norm_layer,
            )
        )
        self.num_inchannels[branch_index] = (
            num_channels[branch_index] * block.expansion
        )
        for i in range(1, num_blocks[branch_index]):
            layers.append(
                block(
                    self.num_inchannels[branch_index],
                    num_channels[branch_index],
                    norm_layer=self.norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _make_branches(self, num_branches, block, num_blocks, num_channels):
        branches = []

        for i in range(num_branches):
            branches.append(
                self._make_one_branch(i, block, num_blocks, num_channels)
            )

        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        num_inchannels = self.num_inchannels
        fuse_layers = []
        for i in range(num_branches if self.multi_scale_output else 1):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    fuse_layer.append(
                        nn.Sequential(
                            nn.Conv2d(
                                in_channels=num_inchannels[j],
                                out_channels=num_inchannels[i],
                                kernel_size=1,
                                bias=False,
                            ),
                            self.norm_layer(num_inchannels[i]),
                        )
                    )
                elif j == i:
                    fuse_layer.append(None)
                else:
                    conv3x3s = []
                    for k in range(i - j):
                        if k == i - j - 1:
                            num_outchannels_conv3x3 = num_inchannels[i]
                            conv3x3s.append(
                                nn.Sequential(
                                    nn.Conv2d(
                                        num_inchannels[j],
                                        num_outchannels_conv3x3,
                                        kernel_size=3,
                                        stride=2,
                                        padding=1,
                                        bias=False,
                                    ),
                                    self.norm_layer(num_outchannels_conv3x3),
                                )
                            )
                        else:
                            num_outchannels_conv3x3 = num_inchannels[j]
                            conv3x3s.append(
                                nn.Sequential(
                                    nn.Conv2d(
                                        num_inchannels[j],
                                        num_outchannels_conv3x3,
                                        kernel_size=3,
                                        stride=2,
                                        padding=1,
                                        bias=False,
                                    ),
                                    self.norm_layer(num_outchannels_conv3x3),
                                    nn.ReLU(inplace=relu_inplace),
                                )
                            )
                    fuse_layer.append(nn.Sequential(*conv3x3s))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)

    def get_num_inchannels(self):
        return self.num_inchannels

    def forward(self, x):
        if self.num_branches == 1:
            return [self.branches[0](x[0])]

        for i in range(self.num_branches):
            x[i] = self.branches[i](x[i])

        x_fuse = []
        for i in range(len(self.fuse_layers)):
            y = x[0] if i == 0 else self.fuse_layers[i][0](x[0])
            for j in range(1, self.num_branches):
                if i == j:
                    y = y + x[j]
                elif j > i:
                    width_output = x[i].shape[-1]
                    height_output = x[i].shape[-2]
                    y = y + F.interpolate(
                        self.fuse_layers[i][j](x[j]),
                        size=[height_output, width_output],
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                else:
                    y = y + self.fuse_layers[i][j](x[j])
            x_fuse.append(self.relu(y))

        return x_fuse


class HighResolutionNet(nn.Module):
    def __init__(
        self,
        width,
        num_classes,
        ocr_width=256,
        small=False,
        norm_layer=nn.BatchNorm2d,
        align_corners=True,
    ):
        super(HighResolutionNet, self).__init__()
        self.norm_layer = norm_layer
        self.width = width
        self.ocr_width = ocr_width
        self.align_corners = align_corners

        self.conv1 = nn.Conv2d(
            3, 64, kernel_size=3, stride=2, padding=1, bias=False
        )
        self.bn1 = norm_layer(64)
        self.conv2 = nn.Conv2d(
            64, 64, kernel_size=3, stride=2, padding=1, bias=False
        )
        self.bn2 = norm_layer(64)
        self.relu = nn.ReLU(inplace=relu_inplace)

        num_blocks = 2 if small else 4

        stage1_num_channels = 64
        self.layer1 = self._make_layer(
            BottleneckV1b, 64, stage1_num_channels, blocks=num_blocks
        )
        stage1_out_channel = BottleneckV1b.expansion * stage1_num_channels

        self.stage2_num_branches = 2
        num_channels = [width, 2 * width]
        num_inchannels = [
            num_channels[i] * BasicBlockV1b.expansion
            for i in range(len(num_channels))
        ]
        self.transition1 = self._make_transition_layer(
            [stage1_out_channel], num_inchannels
        )
        self.stage2, pre_stage_channels = self._make_stage(
            BasicBlockV1b,
            num_inchannels=num_inchannels,
            num_modules=1,
            num_branches=self.stage2_num_branches,
            num_blocks=2 * [num_blocks],
            num_channels=num_channels,
        )

        self.stage3_num_branches = 3
        num_channels = [width, 2 * width, 4 * width]
        num_inchannels = [
            num_channels[i] * BasicBlockV1b.expansion
            for i in range(len(num_channels))
        ]
        self.transition2 = self._make_transition_layer(
            pre_stage_channels, num_inchannels
        )
        self.stage3, pre_stage_channels = self._make_stage(
            BasicBlockV1b,
            num_inchannels=num_inchannels,
            num_modules=3 if small else 4,
            num_branches=self.stage3_num_branches,
            num_blocks=3 * [num_blocks],
            num_channels=num_channels,
        )

        self.stage4_num_branches = 4
        num_channels = [width, 2 * width, 4 * width, 8 * width]
        num_inchannels = [
            num_channels[i] * BasicBlockV1b.expansion
            for i in range(len(num_channels))
        ]
        self.transition3 = self._make_transition_layer(
            pre_stage_channels, num_inchannels
        )
        self.stage4, pre_stage_channels = self._make_stage(
            BasicBlockV1b,
            num_inchannels=num_inchannels,
            num_modules=2 if small else 3,
            num_branches=self.stage4_num_branches,
            num_blocks=4 * [num_blocks],
            num_channels=num_channels,
        )

        last_inp_channels = int(np.sum(pre_stage_channels))
        if self.ocr_width > 0:
            ocr_mid_channels = 2 * self.ocr_width
            ocr_key_channels = self.ocr_width

            self.conv3x3_ocr = nn.Sequential(
                nn.Conv2d(
                    last_inp_channels,
                    ocr_mid_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                ),
                norm_layer(ocr_mid_channels),
                nn.ReLU(inplace=relu_inplace),
            )
            self.ocr_gather_head = SpatialGather_Module(num_classes)

            self.ocr_distri_head = SpatialOCR_Module(
                in_channels=ocr_mid_channels,
                key_channels=ocr_key_channels,
                out_channels=ocr_mid_channels,
                scale=1,
                dropout=0.05,
                norm_layer=norm_layer,
                align_corners=align_corners,
            )
            self.cls_head = nn.Conv2d(
                ocr_mid_channels,
                num_classes,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=True,
            )

            self.aux_head = nn.Sequential(
                nn.Conv2d(
                    last_inp_channels,
                    last_inp_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                ),
                norm_layer(last_inp_channels),
                nn.ReLU(inplace=relu_inplace),
                nn.Conv2d(
                    last_inp_channels,
                    num_classes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=True,
                ),
            )
        else:
            self.cls_head = nn.Sequential(
                nn.Conv2d(
                    last_inp_channels,
                    last_inp_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                ),
                norm_layer(last_inp_channels),
                nn.ReLU(inplace=relu_inplace),
                nn.Conv2d(
                    last_inp_channels,
                    num_classes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=True,
                ),
            )

    def _make_transition_layer(
        self, num_channels_pre_layer, num_channels_cur_layer
    ):
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(
                        nn.Sequential(
                            nn.Conv2d(
                                num_channels_pre_layer[i],
                                num_channels_cur_layer[i],
                                kernel_size=3,
                                stride=1,
                                padding=1,
                                bias=False,
                            ),
                            self.norm_layer(num_channels_cur_layer[i]),
                            nn.ReLU(inplace=relu_inplace),
                        )
                    )
                else:
                    transition_layers.append(None)
            else:
                conv3x3s = []
                for j in range(i + 1 - num_branches_pre):
                    inchannels = num_channels_pre_layer[-1]
                    outchannels = (
                        num_channels_cur_layer[i]
                        if j == i - num_branches_pre
                        else inchannels
                    )
                    conv3x3s.append(
                        nn.Sequential(
                            nn.Conv2d(
                                inchannels,
                                outchannels,
                                kernel_size=3,
                                stride=2,
                                padding=1,
                                bias=False,
                            ),
                            self.norm_layer(outchannels),
                            nn.ReLU(inplace=relu_inplace),
                        )
                    )
                transition_layers.append(nn.Sequential(*conv3x3s))

        return nn.ModuleList(transition_layers)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                self.norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                inplanes,
                planes,
                stride,
                downsample=downsample,
                norm_layer=self.norm_layer,
            )
        )
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes, norm_layer=self.norm_layer))

        return nn.Sequential(*layers)

    def _make_stage(
        self,
        block,
        num_inchannels,
        num_modules,
        num_branches,
        num_blocks,
        num_channels,
        fuse_method="SUM",
        multi_scale_output=True,
    ):
        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            if not multi_scale_output and i == num_modules - 1:
                reset_multi_scale_output = False
            else:
                reset_multi_scale_output = True
            modules.append(
                HighResolutionModule(
                    num_branches,
                    block,
                    num_blocks,
                    num_inchannels,
                    num_channels,
                    fuse_method,
                    reset_multi_scale_output,
                    norm_layer=self.norm_layer,
                    align_corners=self.align_corners,
                )
            )
            num_inchannels = modules[-1].get_num_inchannels()

        return nn.Sequential(*modules), num_inchannels

    def forward(self, x, additional_features=None):
        feats = self.compute_hrnet_feats(x, additional_features)
        if self.ocr_width > 0:
            out_aux = self.aux_head(feats)
            feats = self.conv3x3_ocr(feats)

            context = self.ocr_gather_head(feats, out_aux)
            feats = self.ocr_distri_head(feats, context)
            out = self.cls_head(feats)
            return [out, out_aux]
        else:
            return [self.cls_head(feats), None]

    def compute_hrnet_feats(self, x, additional_features):
        x = self.compute_pre_stage_features(x, additional_features)
        x = self.layer1(x)

        x_list = []
        for i in range(self.stage2_num_branches):
            if self.transition1[i] is not None:
                x_list.append(self.transition1[i](x))
            else:
                x_list.append(x)
        y_list = self.stage2(x_list)

        x_list = []
        for i in range(self.stage3_num_branches):
            if self.transition2[i] is not None:
                if i < self.stage2_num_branches:
                    x_list.append(self.transition2[i](y_list[i]))
                else:
                    x_list.append(self.transition2[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage3(x_list)

        x_list = []
        for i in range(self.stage4_num_branches):
            if self.transition3[i] is not None:
                if i < self.stage3_num_branches:
                    x_list.append(self.transition3[i](y_list[i]))
                else:
                    x_list.append(self.transition3[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        x = self.stage4(x_list)

        return self.aggregate_hrnet_features(x)

    def compute_pre_stage_features(self, x, additional_features):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        if additional_features is not None:
            x = x + additional_features
        x = self.conv2(x)
        x = self.bn2(x)
        return self.relu(x)

    def aggregate_hrnet_features(self, x):
        # Upsampling
        x0_h, x0_w = x[0].size(2), x[0].size(3)
        x1 = F.interpolate(
            x[1],
            size=(x0_h, x0_w),
            mode="bilinear",
            align_corners=self.align_corners,
        )
        x2 = F.interpolate(
            x[2],
            size=(x0_h, x0_w),
            mode="bilinear",
            align_corners=self.align_corners,
        )
        x3 = F.interpolate(
            x[3],
            size=(x0_h, x0_w),
            mode="bilinear",
            align_corners=self.align_corners,
        )

        return torch.cat([x[0], x1, x2, x3], 1)

    def load_pretrained_weights(self, pretrained_path=""):
        model_dict = self.state_dict()

        if not os.path.exists(pretrained_path):
            print(f'\nFile "{pretrained_path}" does not exist.')
            print(
                "You need to specify the correct path to the pre-trained weights.\n"  # noqa: E501
                "You can download the weights for HRNet from the repository:\n"
                "https://github.com/HRNet/HRNet-Image-Classification"
            )
            exit(1)
        pretrained_dict = torch.load(
            pretrained_path, map_location={"cuda:0": "cpu"}
        )
        pretrained_dict = {
            k.replace("last_layer", "aux_head").replace("model.", ""): v
            for k, v in pretrained_dict.items()
        }

        pretrained_dict = {
            k: v for k, v in pretrained_dict.items() if k in model_dict.keys()
        }

        model_dict.update(pretrained_dict)
        self.load_state_dict(model_dict)



import importlib
import logging

import numpy as np
import torch

logger = logging.getLogger(__name__)


def get_dims_with_exclusion(dim, exclude=None):
    dims = list(range(dim))
    if exclude is not None:
        dims.remove(exclude)

    return dims


def save_checkpoint(
    net, checkpoints_path, epoch=None, prefix="", multi_gpu=False
):
    if epoch is None:
        checkpoint_name = "last_checkpoint.pth"
    else:
        checkpoint_name = f"{epoch:03d}.pth"

    if prefix:
        checkpoint_name = f"{prefix}_{checkpoint_name}"

    if not checkpoints_path.exists():
        checkpoints_path.mkdir(parents=True)

    checkpoint_path = checkpoints_path / checkpoint_name
    logger.debug(f"Save checkpoint to {str(checkpoint_path)}")

    net = net.module if multi_gpu else net
    torch.save(
        {"state_dict": net.state_dict(), "config": net._config},
        str(checkpoint_path),
    )


def get_bbox_from_mask(mask):
    rows = np.any(mask, axis=1)
    cols = np.any(mask, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def expand_bbox(bbox, expand_ratio, min_crop_size=None):
    rmin, rmax, cmin, cmax = bbox
    rcenter = 0.5 * (rmin + rmax)
    ccenter = 0.5 * (cmin + cmax)
    height = expand_ratio * (rmax - rmin + 1)
    width = expand_ratio * (cmax - cmin + 1)
    if min_crop_size is not None:
        height = max(height, min_crop_size)
        width = max(width, min_crop_size)

    rmin = int(round(rcenter - 0.5 * height))
    rmax = int(round(rcenter + 0.5 * height))
    cmin = int(round(ccenter - 0.5 * width))
    cmax = int(round(ccenter + 0.5 * width))

    return rmin, rmax, cmin, cmax


def clamp_bbox(bbox, rmin, rmax, cmin, cmax):
    return (
        max(rmin, bbox[0]),
        min(rmax, bbox[1]),
        max(cmin, bbox[2]),
        min(cmax, bbox[3]),
    )


def get_bbox_iou(b1, b2):
    h_iou = get_segments_iou(b1[:2], b2[:2])
    w_iou = get_segments_iou(b1[2:4], b2[2:4])
    return h_iou * w_iou


def get_segments_iou(s1, s2):
    a, b = s1
    c, d = s2
    intersection = max(0, min(b, d) - max(a, c) + 1)
    union = max(1e-6, max(b, d) - min(a, c) + 1)
    return intersection / union


def get_labels_with_sizes(x):
    obj_sizes = np.bincount(x.flatten())
    labels = np.nonzero(obj_sizes)[0].tolist()
    labels = [x for x in labels if x != 0]
    return labels, obj_sizes[labels].tolist()


def ignore_params_then_call(func):
    def ret(*args, **kwargs):
        return func()

    return ret


def load_module(script_path, module_name="module"):
    logger.debug(f"Loading module '{script_path}'...")
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    assert spec is not None, f"Can't import model at '{script_path}'"
    model_script = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(model_script)

    return model_script



import tkinter as tk
from tkinter import messagebox, ttk


class BoundedNumericalEntry(tk.Entry):
    def __init__(
        self,
        master=None,
        min_value=None,
        max_value=None,
        variable=None,
        vartype=float,
        width=7,
        allow_inf=False,
        **kwargs,
    ):
        if variable is None:
            if vartype == float:
                self.var = tk.DoubleVar()
            elif vartype == int:
                self.var = tk.IntVar()
            else:
                self.var = tk.StringVar()
        else:
            self.var = variable

        self.fake_var = tk.StringVar(value=self.var.get())
        self.vartype = vartype
        self.old_value = self.var.get()
        self.allow_inf = allow_inf

        self.min_value, self.max_value = min_value, max_value
        self.get, self.set = self.fake_var.get, self.fake_var.set

        self.validate_command = master.register(self._check_bounds)
        tk.Entry.__init__(
            self,
            master,
            textvariable=self.fake_var,
            validate="focus",
            width=width,
            vcmd=(self.validate_command, "%P", "%d"),
            **kwargs,
        )

    def _check_bounds(self, instr, action_type):
        if self.allow_inf and instr == "INF":
            self.fake_var.set("INF")
            return True

        if action_type == "-1":
            try:
                new_value = self.vartype(instr)
            except ValueError:
                pass
            else:
                if (
                    self.min_value is None or new_value >= self.min_value
                ) and (self.max_value is None or new_value <= self.max_value):
                    if new_value != self.old_value:
                        self.old_value = self.vartype(self.fake_var.get())
                        self.delete(0, tk.END)
                        self.insert(0, str(self.old_value))
                        self.var.set(self.old_value)
                    return True
        self.delete(0, tk.END)
        self.insert(0, str(self.old_value))
        mn = "-inf" if self.min_value is None else str(self.min_value)
        mx = "+inf" if self.max_value is None else str(self.max_value)
        messagebox.showwarning(
            "Incorrect value in input field",
            f"Value for {self._name} should be in "
            f"[{mn}; {mx}] and of type {self.vartype.__name__}",
        )

        return False


class FocusHorizontalScale(tk.Scale):
    def __init__(
        self,
        *args,
        highlightthickness=0,
        sliderrelief=tk.GROOVE,
        resolution=0.01,
        sliderlength=20,
        length=200,
        **kwargs,
    ):
        tk.Scale.__init__(
            self,
            *args,
            orient=tk.HORIZONTAL,
            highlightthickness=highlightthickness,
            sliderrelief=sliderrelief,
            resolution=resolution,
            sliderlength=sliderlength,
            length=length,
            **kwargs,
        )
        self.bind("<1>", lambda event: self.focus_set())


class FocusCheckButton(tk.Checkbutton):
    def __init__(self, *args, highlightthickness=0, **kwargs):
        tk.Checkbutton.__init__(
            self, *args, highlightthickness=highlightthickness, **kwargs
        )
        self.bind("<1>", lambda event: self.focus_set())


class FocusButton(tk.Button):
    def __init__(self, *args, highlightthickness=0, **kwargs):
        tk.Button.__init__(
            self, *args, highlightthickness=highlightthickness, **kwargs
        )
        self.bind("<1>", lambda event: self.focus_set())


class FocusLabelFrame(ttk.LabelFrame):
    def __init__(
        self,
        *args,
        highlightthickness=0,
        relief=tk.RIDGE,
        borderwidth=2,
        **kwargs,
    ):
        tk.LabelFrame.__init__(
            self,
            *args,
            highlightthickness=highlightthickness,
            relief=relief,
            borderwidth=borderwidth,
            **kwargs,
        )
        self.bind("<1>", lambda event: self.focus_set())

    def set_frame_state(self, state):
        def set_widget_state(widget, state):
            if widget.winfo_children is not None:
                for w in widget.winfo_children():
                    w.configure(state=state)

        set_widget_state(self, state)



import pickle as pkl
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class PascalVocDataset(ISDataset):
    def __init__(self, dataset_path, split="train", dry_run=False, **kwargs):
        super().__init__(**kwargs)
        assert split in {"train", "val", "trainval", "test"}

        self.dataset_path = Path(dataset_path)
        self._images_path = self.dataset_path / "JPEGImages"
        self._insts_path = self.dataset_path / "SegmentationObject"
        self.dataset_split = split
        if dry_run:
            self.dataset_samples = []
        else:
            if split == "test":
                with open(
                    self.dataset_path / "ImageSets/Segmentation/test.pickle",
                    "rb",
                ) as f:
                    self.dataset_samples, self.instance_ids = pkl.load(f)
            else:
                with open(
                    self.dataset_path / f"ImageSets/Segmentation/{split}.txt",
                    "r",
                ) as f:
                    self.dataset_samples = [
                        name.strip() for name in f.readlines()
                    ]

    def get_sample(self, index) -> DSample:
        sample_id = self.dataset_samples[index]
        image_path = str(self._images_path / f"{sample_id}.jpg")
        mask_path = str(self._insts_path / f"{sample_id}.png")

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = cv2.imread(mask_path)
        instances_mask = cv2.cvtColor(
            instances_mask, cv2.COLOR_BGR2GRAY
        ).astype(np.int32)
        if self.dataset_split == "test":
            instance_id = self.instance_ids[index]
            mask = np.zeros_like(instances_mask)
            mask[instances_mask == 220] = 220  # ignored area
            mask[instances_mask == instance_id] = 1
            objects_ids = [1]
            instances_mask = mask
        else:
            objects_ids = np.unique(instances_mask)  # type: ignore
            objects_ids = [x for x in objects_ids if x != 0 and x != 220]

        return DSample(
            image,
            instances_mask,
            objects_ids=objects_ids,
            ignore_ids=[220],
            sample_id=index,
        )



# -*- coding: utf-8 -*-
""" Adopted from https://github.com/foobar167/junkyard/blob/master/manual_image_annotation1/polygon/gui_canvas.py """  # noqa: E501
import math
import os
import sys
import time
import tkinter as tk
from tkinter import ttk

from PIL import Image, ImageTk


def handle_exception(exit_code=0):
    """Use: @land.logger.handle_exception(0)
    before every function which could cast an exception"""

    def wrapper(func):
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:  # noqa: E722
                if exit_code != 0:  # if zero, don't exit from the program
                    sys.exit(exit_code)  # exit from the program

        return inner

    return wrapper


class AutoScrollbar(ttk.Scrollbar):
    """A scrollbar that hides itself if it's not needed. Works only for grid geometry manager"""  # noqa: E501

    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
            ttk.Scrollbar.set(self, lo, hi)

    @handle_exception(1)
    def pack(self, **kw):
        raise tk.TclError(
            "Cannot use pack with the widget " + self.__class__.__name__
        )

    @handle_exception(1)
    def place(self, **kw):
        raise tk.TclError(
            "Cannot use place with the widget " + self.__class__.__name__
        )


class CanvasImage:
    """Display and zoom image"""

    def __init__(self, canvas_frame, canvas):
        """Initialize the ImageFrame"""
        self.current_scale = (
            1.0  # scale for the canvas image zoom, public for outer classes
        )
        self.__delta = 1.2  # zoom magnitude
        self.__previous_state = 0  # previous state of the keyboard
        # Create ImageFrame in placeholder widget
        self.__imframe = canvas_frame
        # Vertical and horizontal scrollbars for canvas
        self.hbar = AutoScrollbar(canvas_frame, orient="horizontal")
        self.vbar = AutoScrollbar(canvas_frame, orient="vertical")
        self.hbar.grid(row=1, column=0, sticky="we")
        self.vbar.grid(row=0, column=1, sticky="ns")
        # Add scroll bars to canvas
        self.canvas = canvas
        self.canvas.configure(
            xscrollcommand=self.hbar.set, yscrollcommand=self.vbar.set
        )
        self.hbar.configure(
            command=self.__scroll_x
        )  # bind scrollbars to the canvas
        self.vbar.configure(command=self.__scroll_y)
        # Bind events to the Canvas
        self.canvas.bind(
            "<Configure>", lambda event: self.__size_changed()
        )  # canvas is resized
        self.canvas.bind(
            "<Button-1>", self.__left_mouse_button
        )  # remember canvas position
        self.canvas.bind(
            "<ButtonPress-3>", self.__right_mouse_button_pressed
        )  # remember canvas position
        self.canvas.bind(
            "<ButtonPress-2>", self.__right_mouse_button_pressed
        )  # remember canvas position (MacOS)
        self.canvas.bind(
            "<ButtonRelease-3>", self.__right_mouse_button_released
        )  # remember canvas position
        self.canvas.bind(
            "<ButtonRelease-2>", self.__right_mouse_button_released
        )  # remember canvas position (MacOS)
        self.canvas.bind(
            "<B3-Motion>", self.__right_mouse_button_motion
        )  # move canvas to the new position
        self.canvas.bind(
            "<B2-Motion>", self.__right_mouse_button_motion
        )  # move canvas to the new position
        self.canvas.bind(
            "<MouseWheel>", self.__wheel
        )  # zoom for Windows and MacOS, but not Linux
        self.canvas.bind(
            "<Button-5>", self.__wheel
        )  # zoom for Linux, wheel scroll down
        self.canvas.bind(
            "<Button-4>", self.__wheel
        )  # zoom for Linux, wheel scroll up

        # Handle keystrokes in idle mode, because program slows down on a weak computers,  # noqa: E501
        # when too many key stroke events in the same time
        self.canvas.bind(
            "<Key>",
            lambda event: self.canvas.after_idle(self.__keystroke, event),
        )
        self.container = None

        self._click_callback = None

    def register_click_callback(self, click_callback):
        self._click_callback = click_callback

    def reload_image(self, image, reset_canvas=True):
        self.__original_image = image.copy()
        self.__current_image = image.copy()

        if reset_canvas:
            self.imwidth, self.imheight = self.__original_image.size
            self.__min_side = min(
                self.imwidth, self.imheight
            )  # get the smaller image side

            scale = min(
                self.canvas.winfo_width() / self.imwidth,
                self.canvas.winfo_height() / self.imheight,
            )
            if self.container:
                self.canvas.delete(self.container)

            self.container = self.canvas.create_rectangle(
                (0, 0, scale * self.imwidth, scale * self.imheight), width=0
            )
            self.current_scale = scale
            self._reset_canvas_offset()

        self.__show_image()  # show image on the canvas
        self.canvas.focus_set()  # set focus on the canvas

    def grid(self, **kw):
        """Put CanvasImage widget on the parent widget"""
        self.__imframe.grid(**kw)  # place CanvasImage widget on the grid
        self.__imframe.grid(sticky="nswe")  # make frame container sticky
        self.__imframe.rowconfigure(0, weight=1)  # make canvas expandable
        self.__imframe.columnconfigure(0, weight=1)

    def __show_image(self):
        box_image = self.canvas.coords(self.container)  # get image area
        box_canvas = (
            self.canvas.canvasx(0),  # get visible area of the canvas
            self.canvas.canvasy(0),
            self.canvas.canvasx(self.canvas.winfo_width()),
            self.canvas.canvasy(self.canvas.winfo_height()),
        )
        box_img_int = tuple(
            map(int, box_image)
        )  # convert to integer or it will not work properly
        # Get scroll region box
        box_scroll = [
            min(box_img_int[0], box_canvas[0]),
            min(box_img_int[1], box_canvas[1]),
            max(box_img_int[2], box_canvas[2]),
            max(box_img_int[3], box_canvas[3]),
        ]
        # Horizontal part of the image is in the visible area
        if box_scroll[0] == box_canvas[0] and box_scroll[2] == box_canvas[2]:
            box_scroll[0] = box_img_int[0]
            box_scroll[2] = box_img_int[2]
        # Vertical part of the image is in the visible area
        if box_scroll[1] == box_canvas[1] and box_scroll[3] == box_canvas[3]:
            box_scroll[1] = box_img_int[1]
            box_scroll[3] = box_img_int[3]
        # Convert scroll region to tuple and to integer
        self.canvas.configure(
            scrollregion=tuple(map(int, box_scroll))
        )  # set scroll region
        x1 = max(
            box_canvas[0] - box_image[0], 0
        )  # get coordinates (x1,y1,x2,y2) of the image tile
        y1 = max(box_canvas[1] - box_image[1], 0)
        x2 = min(box_canvas[2], box_image[2]) - box_image[0]
        y2 = min(box_canvas[3], box_image[3]) - box_image[1]

        if (
            int(x2 - x1) > 0 and int(y2 - y1) > 0
        ):  # show image if it in the visible area
            border_width = 2
            sx1, sx2 = x1 / self.current_scale, x2 / self.current_scale
            sy1, sy2 = y1 / self.current_scale, y2 / self.current_scale
            crop_x, crop_y = max(0, math.floor(sx1 - border_width)), max(
                0, math.floor(sy1 - border_width)
            )
            crop_w, crop_h = math.ceil(
                sx2 - sx1 + 2 * border_width
            ), math.ceil(sy2 - sy1 + 2 * border_width)
            crop_w = min(crop_w, self.__original_image.width - crop_x)
            crop_h = min(crop_h, self.__original_image.height - crop_y)

            __current_image = self.__original_image.crop(
                (crop_x, crop_y, crop_x + crop_w, crop_y + crop_h)
            )
            crop_zw = int(round(crop_w * self.current_scale))
            crop_zh = int(round(crop_h * self.current_scale))
            zoom_sx, zoom_sy = crop_zw / crop_w, crop_zh / crop_h
            crop_zx, crop_zy = crop_x * zoom_sx, crop_y * zoom_sy
            self.real_scale = (zoom_sx, zoom_sy)

            interpolation = (
                Image.NEAREST if self.current_scale > 2.0 else Image.ANTIALIAS
            )
            __current_image = __current_image.resize(
                (crop_zw, crop_zh), interpolation
            )
            zx1, zy1 = x1 - crop_zx, y1 - crop_zy
            zx2 = min(zx1 + self.canvas.winfo_width(), __current_image.width)
            zy2 = min(zy1 + self.canvas.winfo_height(), __current_image.height)

            self.__current_image = __current_image.crop((zx1, zy1, zx2, zy2))

            imagetk = ImageTk.PhotoImage(self.__current_image)
            imageid = self.canvas.create_image(
                max(box_canvas[0], box_img_int[0]),
                max(box_canvas[1], box_img_int[1]),
                anchor="nw",
                image=imagetk,
            )
            self.canvas.lower(imageid)  # set image into background
            self.canvas.imagetk = imagetk  # keep an extra reference to prevent garbage-collection # noqa: E501

    def _get_click_coordinates(self, event):
        x = self.canvas.canvasx(
            event.x
        )  # get coordinates of the event on the canvas
        y = self.canvas.canvasy(event.y)

        if self.outside(x, y):
            return None

        box_image = self.canvas.coords(self.container)
        x = max(x - box_image[0], 0)
        y = max(y - box_image[1], 0)

        x = int(x / self.real_scale[0])
        y = int(y / self.real_scale[1])

        return x, y

    # ================================================ Canvas Routines =================================================  # noqa: E501
    def _reset_canvas_offset(self):
        self.canvas.configure(scrollregion=(0, 0, 5000, 5000))
        self.canvas.scan_mark(0, 0)
        self.canvas.scan_dragto(
            int(self.canvas.canvasx(0)), int(self.canvas.canvasy(0)), gain=1
        )

    def _change_canvas_scale(self, relative_scale, x=0, y=0):
        new_scale = self.current_scale * relative_scale

        if new_scale > 20:
            return

        if (
            new_scale * self.__original_image.width < self.canvas.winfo_width()
            and new_scale * self.__original_image.height
            < self.canvas.winfo_height()
        ):
            return

        self.current_scale = new_scale
        self.canvas.scale(
            "all", x, y, relative_scale, relative_scale
        )  # rescale all objects

    # noinspection PyUnusedLocal
    def __scroll_x(self, *args, **kwargs):
        """Scroll canvas horizontally and redraw the image"""
        self.canvas.xview(*args)  # scroll horizontally
        self.__show_image()  # redraw the image

    # noinspection PyUnusedLocal
    def __scroll_y(self, *args, **kwargs):
        """Scroll canvas vertically and redraw the image"""
        self.canvas.yview(*args)  # scroll vertically
        self.__show_image()  # redraw the image

    def __size_changed(self):
        new_scale_w = self.canvas.winfo_width() / (
            self.current_scale * self.__original_image.width
        )
        new_scale_h = self.canvas.winfo_height() / (
            self.current_scale * self.__original_image.height
        )
        new_scale = min(new_scale_w, new_scale_h)
        if new_scale > 1.0:
            self._change_canvas_scale(new_scale)
        self.__show_image()

    # ================================================ Mouse callbacks =================================================  # noqa: E501
    def __wheel(self, event):
        """Zoom with mouse wheel"""
        x = self.canvas.canvasx(
            event.x
        )  # get coordinates of the event on the canvas
        y = self.canvas.canvasy(event.y)
        if self.outside(x, y):
            return  # zoom only inside image area

        scale = 1.0
        # Respond to Linux (event.num) or Windows (event.delta) wheel event
        if (
            event.num == 5 or event.delta == -120 or event.delta == 1
        ):  # scroll down, zoom out, smaller
            scale /= self.__delta
        if (
            event.num == 4 or event.delta == 120 or event.delta == -1
        ):  # scroll up, zoom in, bigger
            scale *= self.__delta

        self._change_canvas_scale(scale, x, y)
        self.__show_image()

    def __left_mouse_button(self, event):
        if self._click_callback is None:
            return

        coords = self._get_click_coordinates(event)

        if coords is not None:
            self._click_callback(is_positive=True, x=coords[0], y=coords[1])

    def __right_mouse_button_pressed(self, event):
        """Remember previous coordinates for scrolling with the mouse"""
        self._last_rb_click_time = time.time()
        self._last_rb_click_event = event
        self.canvas.scan_mark(event.x, event.y)

    def __right_mouse_button_released(self, event):
        time_delta = time.time() - self._last_rb_click_time
        move_delta = math.sqrt(
            (event.x - self._last_rb_click_event.x) ** 2
            + (event.y - self._last_rb_click_event.y) ** 2
        )
        if time_delta > 0.5 or move_delta > 3:
            return

        if self._click_callback is None:
            return

        coords = self._get_click_coordinates(self._last_rb_click_event)

        if coords is not None:
            self._click_callback(is_positive=False, x=coords[0], y=coords[1])

    def __right_mouse_button_motion(self, event):
        """Drag (move) canvas to the new position"""
        move_delta = math.sqrt(
            (event.x - self._last_rb_click_event.x) ** 2
            + (event.y - self._last_rb_click_event.y) ** 2
        )
        if move_delta > 3:
            self.canvas.scan_dragto(event.x, event.y, gain=1)
            self.__show_image()  # zoom tile and show it on the canvas

    def outside(self, x, y):
        """Checks if the point (x,y) is outside the image area"""
        bbox = self.canvas.coords(self.container)  # get image area
        if bbox[0] < x < bbox[2] and bbox[1] < y < bbox[3]:
            return False  # point (x,y) is inside the image area
        else:
            return True  # point (x,y) is outside the image area

    # ================================================= Keys Callback ==================================================  # noqa: E501
    def __keystroke(self, event):
        """Scrolling with the keyboard.
        Independent from the language of the keyboard, CapsLock, <Ctrl>+<key>, etc.
        """  # noqa: E501
        if (
            event.state - self.__previous_state == 4
        ):  # means that the Control key is pressed
            pass  # do nothing if Control key is pressed
        else:
            self.__previous_state = (
                event.state
            )  # remember the last keystroke state
            # Up, Down, Left, Right keystrokes
            self.keycodes = {}  # init key codes
            if os.name == "nt":  # Windows OS
                self.keycodes = {
                    "d": [68, 39, 102],
                    "a": [65, 37, 100],
                    "w": [87, 38, 104],
                    "s": [83, 40, 98],
                }
            else:  # Linux OS
                self.keycodes = {
                    "d": [40, 114, 85],
                    "a": [38, 113, 83],
                    "w": [25, 111, 80],
                    "s": [39, 116, 88],
                }
            if (
                event.keycode in self.keycodes["d"]
            ):  # scroll right, keys 'd' or 'Right'
                self.__scroll_x("scroll", 1, "unit", event=event)
            elif (
                event.keycode in self.keycodes["a"]
            ):  # scroll left, keys 'a' or 'Left'
                self.__scroll_x("scroll", -1, "unit", event=event)
            elif (
                event.keycode in self.keycodes["w"]
            ):  # scroll up, keys 'w' or 'Up'
                self.__scroll_y("scroll", -1, "unit", event=event)
            elif (
                event.keycode in self.keycodes["s"]
            ):  # scroll down, keys 's' or 'Down'
                self.__scroll_y("scroll", 1, "unit", event=event)



from easydict import EasyDict as edict

from .env import load_cfg_from_env


def test_load_cfg_from_env():
    input_dict = {"RITM_a": 2, "RITM_eoq__trabson": 3}
    loaded = load_cfg_from_env(edict(), input_dict)
    assert loaded.a == 2
    assert loaded.eoq.trabson == 3



class LRMult(object):
    def __init__(self, lr_mult=1.0):
        self.lr_mult = lr_mult

    def __call__(self, m):
        if getattr(m, "weight", None) is not None:
            m.weight.lr_mult = self.lr_mult
        if getattr(m, "bias", None) is not None:
            m.bias.lr_mult = self.lr_mult



from .grabcut import GrabCutDataset


class BerkeleyDataset(GrabCutDataset):
    def __init__(self, dataset_path, **kwargs):
        super().__init__(
            dataset_path,
            images_dir_name="images",
            masks_dir_name="masks",
            **kwargs
        )



import numpy as np
import torch
from torch import nn as nn

import ritm_annotation.model.initializer as initializer


def select_activation_function(activation):
    if isinstance(activation, str):
        if activation.lower() == "relu":
            return nn.ReLU
        elif activation.lower() == "softplus":
            return nn.Softplus
        else:
            raise ValueError(f"Unknown activation type {activation}")
    elif isinstance(activation, nn.Module):
        return activation
    else:
        raise ValueError(f"Unknown activation type {activation}")


class BilinearConvTranspose2d(nn.ConvTranspose2d):
    def __init__(self, in_channels, out_channels, scale, groups=1):
        kernel_size = 2 * scale - scale % 2
        self.scale = scale

        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=scale,
            padding=1,
            groups=groups,
            bias=False,
        )

        self.apply(
            initializer.Bilinear(
                scale=scale, in_channels=in_channels, groups=groups
            )
        )


class DistMaps(nn.Module):
    def __init__(
        self, norm_radius, spatial_scale=1.0, cpu_mode=False, use_disks=False
    ):
        super(DistMaps, self).__init__()
        self.spatial_scale = spatial_scale
        self.norm_radius = norm_radius
        self.cpu_mode = cpu_mode
        self.use_disks = use_disks
        if self.cpu_mode:
            from ritm_annotation.utils.cython import get_dist_maps

            self._get_dist_maps = get_dist_maps

    def get_coord_features(self, points, batchsize, rows, cols):
        if self.cpu_mode:
            coords = []
            for i in range(batchsize):
                norm_delimeter = (
                    1.0
                    if self.use_disks
                    else self.spatial_scale * self.norm_radius
                )
                coords.append(
                    self._get_dist_maps(
                        points[i].cpu().float().numpy(),
                        rows,
                        cols,
                        norm_delimeter,
                    )
                )
            coords = (
                torch.from_numpy(np.stack(coords, axis=0))
                .to(points.device)
                .float()
            )
        else:
            num_points = points.shape[1] // 2
            points = points.view(-1, points.size(2))
            points, points_order = torch.split(points, [2, 1], dim=1)

            invalid_points = torch.max(points, dim=1, keepdim=False)[0] < 0
            row_array = torch.arange(
                start=0,
                end=rows,
                step=1,
                dtype=torch.float32,
                device=points.device,
            )
            col_array = torch.arange(
                start=0,
                end=cols,
                step=1,
                dtype=torch.float32,
                device=points.device,
            )

            coord_rows, coord_cols = torch.meshgrid(row_array, col_array)
            coords = (
                torch.stack((coord_rows, coord_cols), dim=0)
                .unsqueeze(0)
                .repeat(points.size(0), 1, 1, 1)
            )

            add_xy = (points * self.spatial_scale).view(
                points.size(0), points.size(1), 1, 1
            )
            coords.add_(-add_xy)
            if not self.use_disks:
                coords.div_(self.norm_radius * self.spatial_scale)
            coords.mul_(coords)

            coords[:, 0] += coords[:, 1]
            coords = coords[:, :1]

            coords[invalid_points, :, :, :] = 1e6

            coords = coords.view(-1, num_points, 1, rows, cols)
            coords = coords.min(dim=1)[
                0
            ]  # -> (bs * num_masks * 2) x 1 x h x w
            coords = coords.view(-1, 2, rows, cols)

        if self.use_disks:
            coords = (
                coords <= (self.norm_radius * self.spatial_scale) ** 2
            ).float()
        else:
            coords.sqrt_().mul_(2).tanh_()

        return coords

    def forward(self, x, coords):
        return self.get_coord_features(
            coords, x.shape[0], x.shape[2], x.shape[3]
        )


class ScaleLayer(nn.Module):
    def __init__(self, init_value=1.0, lr_mult=1):
        super().__init__()
        self.lr_mult = lr_mult
        self.scale = nn.Parameter(
            torch.full((1,), init_value / lr_mult, dtype=torch.float32)
        )

    def forward(self, x):
        scale = torch.abs(self.scale * self.lr_mult)
        return x * scale


class BatchImageNormalize:
    def __init__(self, mean, std, dtype=torch.float):
        self.mean = torch.as_tensor(mean, dtype=dtype)[None, :, None, None]
        self.std = torch.as_tensor(std, dtype=dtype)[None, :, None, None]

    def __call__(self, tensor):
        tensor = tensor.clone()

        tensor.sub_(self.mean.to(tensor.device)).div_(
            self.std.to(tensor.device)
        )
        return tensor



import logging
import math

import torch

logger = logging.getLogger(__name__)


def get_optimizer(model, opt_name, opt_kwargs):
    params = []
    base_lr = opt_kwargs["lr"]
    for name, param in model.named_parameters():
        param_group = {"params": [param]}
        if not param.requires_grad:
            params.append(param_group)
            continue

        if not math.isclose(getattr(param, "lr_mult", 1.0), 1.0):
            logger.info(
                f'Applied lr_mult={param.lr_mult} to "{name}" parameter.'
            )
            param_group["lr"] = param_group.get("lr", base_lr) * param.lr_mult

        params.append(param_group)

    optimizer = {
        "sgd": torch.optim.SGD,
        "adam": torch.optim.Adam,
        "adamw": torch.optim.AdamW,
    }[opt_name.lower()](params, **opt_kwargs)

    return optimizer



from pathlib import Path

import pytest
from easydict import EasyDict as edict

from ritm_annotation.engine.trainer import ISTrainer
from ritm_annotation.utils.misc import load_module

all_models = []

for model in Path(__file__).parent.glob("**/*.py"):
    if model.name == "__init__.py":
        continue
    if model.name == "test_exposed_model_stuff.py":
        continue
    all_models.append(model)


@pytest.mark.parametrize(
    "model_file",
    all_models,
    ids=[f"{x.parent.name}/{x.name}" for x in all_models],
)
def test_model_exposes_the_right_stuff(model_file):
    model_script = load_module(model_file)

    assert isinstance(model_script.MODEL_NAME, str)
    assert (
        model_script.__dict__.get("main") is None
    ), "Remove the main function"

    cfg = edict()
    cfg.device = "cpu"
    cfg.batch_size = -1  # use default
    cfg.distributed = False
    cfg.workers = 0  # it's dataloader stuff
    cfg.weights = None
    cfg.resume_exp = None
    cfg.multi_gpu = False
    cfg.local_rank = 0  # master
    cfg.start_epoch = 0

    model, model_cfg = model_script.init_model(cfg, dry_run=True)
    trainer = model_script.get_trainer(
        model, cfg, model_cfg, dry_run=True, no_dataset=True
    )

    assert isinstance(model_cfg.default_num_epochs, int)
    assert isinstance(trainer, ISTrainer)
    assert trainer.trainset is None  # test handling of no_dataset
    assert trainer.valset is None



import os
import pickle as pkl
import random
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class OpenImagesDataset(ISDataset):
    def __init__(self, dataset_path, split="train", dry_run=False, **kwargs):
        super().__init__(**kwargs)
        assert split in {"train", "val", "test"}

        self.dataset_path = Path(dataset_path)
        self._split_path = self.dataset_path / split
        self._images_path = self._split_path / "images"
        self._masks_path = self._split_path / "masks"
        self.dataset_split = split

        clean_anno_path = (
            self._split_path
            / f"{split}-annotations-object-segmentation_clean.pkl"
        )
        if os.path.exists(clean_anno_path):
            with clean_anno_path.open("rb") as f:
                annotations = pkl.load(f)
        elif dry_run:
            annotations = dict(image_id_to_masks=[], dataset_samples=[])
        else:
            raise RuntimeError(f"Can't find annotations at {clean_anno_path}")
        self.image_id_to_masks = annotations["image_id_to_masks"]
        self.dataset_samples = annotations["dataset_samples"]

    def get_sample(self, index) -> DSample:
        image_id = self.dataset_samples[index]

        image_path = str(self._images_path / f"{image_id}.jpg")
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        mask_paths = self.image_id_to_masks[image_id]
        # select random mask for an image
        mask_path = str(self._masks_path / random.choice(mask_paths))
        instances_mask = cv2.imread(mask_path)
        instances_mask = cv2.cvtColor(instances_mask, cv2.COLOR_BGR2GRAY)
        instances_mask[instances_mask > 0] = 1
        instances_mask = instances_mask.astype(np.int32)

        min_width = min(image.shape[1], instances_mask.shape[1])
        min_height = min(image.shape[0], instances_mask.shape[0])

        if image.shape[0] != min_height or image.shape[1] != min_width:
            image = cv2.resize(
                image, (min_width, min_height), interpolation=cv2.INTER_LINEAR
            )
        if (
            instances_mask.shape[0] != min_height
            or instances_mask.shape[1] != min_width
        ):
            instances_mask = cv2.resize(
                instances_mask,
                (min_width, min_height),
                interpolation=cv2.INTER_NEAREST,
            )

        object_ids = [1] if instances_mask.sum() > 0 else []

        return DSample(
            image, instances_mask, objects_ids=object_ids, sample_id=index
        )



from ritm_annotation.inference.transforms import ZoomIn
from ritm_annotation.model.is_hrnet_model import HRNetModel

from .base import BasePredictor
from .brs import (
    FeatureBRSPredictor,
    HRNetFeatureBRSPredictor,
    InputBRSPredictor,
)
from .brs_functors import InputOptimizer, ScaleBiasOptimizer


def get_predictor(
    net,
    brs_mode,
    device,
    prob_thresh=0.49,
    with_flip=True,
    zoom_in_params=dict(),
    predictor_params=None,
    brs_opt_func_params=None,
    lbfgs_params=None,
):
    lbfgs_params_ = {
        "m": 20,
        "factr": 0,
        "pgtol": 1e-8,
        "maxfun": 20,
    }

    predictor_params_ = {"optimize_after_n_clicks": 1}

    if zoom_in_params is not None:
        zoom_in = ZoomIn(**zoom_in_params)
    else:
        zoom_in = None

    if lbfgs_params is not None:
        lbfgs_params_.update(lbfgs_params)
    lbfgs_params_["maxiter"] = 2 * lbfgs_params_["maxfun"]

    if brs_opt_func_params is None:
        brs_opt_func_params = dict()

    if isinstance(net, (list, tuple)):
        assert (
            brs_mode == "NoBRS"
        ), "Multi-stage models support only NoBRS mode."

    if brs_mode == "NoBRS":
        if predictor_params is not None:
            predictor_params_.update(predictor_params)
        predictor = BasePredictor(
            net,
            device,
            zoom_in=zoom_in,
            with_flip=with_flip,
            **predictor_params_
        )
    elif brs_mode.startswith("f-BRS"):
        predictor_params_.update(
            {
                "net_clicks_limit": 8,
            }
        )
        if predictor_params is not None:
            predictor_params_.update(predictor_params)

        insertion_mode = {
            "f-BRS-A": "after_c4",
            "f-BRS-B": "after_aspp",
            "f-BRS-C": "after_deeplab",
        }[brs_mode]

        opt_functor = ScaleBiasOptimizer(
            prob_thresh=prob_thresh,
            with_flip=with_flip,
            optimizer_params=lbfgs_params_,
            **brs_opt_func_params
        )

        if isinstance(net, HRNetModel):
            FeaturePredictor = HRNetFeatureBRSPredictor
            insertion_mode = {
                "after_c4": "A",
                "after_aspp": "A",
                "after_deeplab": "C",
            }[insertion_mode]
        else:
            FeaturePredictor = FeatureBRSPredictor

        predictor = FeaturePredictor(
            net,
            device,
            opt_functor=opt_functor,
            with_flip=with_flip,
            insertion_mode=insertion_mode,
            zoom_in=zoom_in,
            **predictor_params_
        )
    elif brs_mode == "RGB-BRS" or brs_mode == "DistMap-BRS":
        use_dmaps = brs_mode == "DistMap-BRS"

        predictor_params_.update(
            {
                "net_clicks_limit": 5,
            }
        )
        if predictor_params is not None:
            predictor_params_.update(predictor_params)

        opt_functor = InputOptimizer(
            prob_thresh=prob_thresh,
            with_flip=with_flip,
            optimizer_params=lbfgs_params_,
            **brs_opt_func_params
        )

        predictor = InputBRSPredictor(
            net,
            device,
            optimize_target="dmaps" if use_dmaps else "rgb",
            opt_functor=opt_functor,
            with_flip=with_flip,
            zoom_in=zoom_in,
            **predictor_params_
        )
    else:
        raise NotImplementedError

    return predictor



from copy import deepcopy

import cv2
import numpy as np


class Clicker(object):
    def __init__(
        self,
        gt_mask=None,
        init_clicks=None,
        ignore_label=-1,
        click_indx_offset=0,
    ):
        self.click_indx_offset = click_indx_offset
        if gt_mask is not None:
            self.gt_mask = gt_mask == 1
            self.not_ignore_mask = gt_mask != ignore_label
        else:
            self.gt_mask = None

        self.reset_clicks()

        if init_clicks is not None:
            for click in init_clicks:
                self.add_click(click)

    def make_next_click(self, pred_mask):
        assert self.gt_mask is not None
        click = self._get_next_click(pred_mask)
        self.add_click(click)

    def get_clicks(self, clicks_limit=None):
        return self.clicks_list[:clicks_limit]

    def _get_next_click(self, pred_mask, padding=True):
        fn_mask = np.logical_and(
            np.logical_and(self.gt_mask, np.logical_not(pred_mask)),
            self.not_ignore_mask,
        )
        fp_mask = np.logical_and(
            np.logical_and(np.logical_not(self.gt_mask), pred_mask),
            self.not_ignore_mask,
        )

        if padding:
            fn_mask = np.pad(fn_mask, ((1, 1), (1, 1)), "constant")
            fp_mask = np.pad(fp_mask, ((1, 1), (1, 1)), "constant")

        fn_mask_dt = cv2.distanceTransform(
            fn_mask.astype(np.uint8), cv2.DIST_L2, 0
        )
        fp_mask_dt = cv2.distanceTransform(
            fp_mask.astype(np.uint8), cv2.DIST_L2, 0
        )

        if padding:
            fn_mask_dt = fn_mask_dt[1:-1, 1:-1]
            fp_mask_dt = fp_mask_dt[1:-1, 1:-1]

        fn_mask_dt = fn_mask_dt * self.not_clicked_map
        fp_mask_dt = fp_mask_dt * self.not_clicked_map

        fn_max_dist = np.max(fn_mask_dt)
        fp_max_dist = np.max(fp_mask_dt)

        is_positive = fn_max_dist > fp_max_dist
        if is_positive:
            coords_y, coords_x = np.where(
                fn_mask_dt == fn_max_dist
            )  # coords is [y, x]
        else:
            coords_y, coords_x = np.where(
                fp_mask_dt == fp_max_dist
            )  # coords is [y, x]

        return Click(
            is_positive=is_positive, coords=(coords_y[0], coords_x[0])
        )

    def add_click(self, click):
        coords = click.coords

        click.indx = (
            self.click_indx_offset + self.num_pos_clicks + self.num_neg_clicks
        )
        if click.is_positive:
            self.num_pos_clicks += 1
        else:
            self.num_neg_clicks += 1

        self.clicks_list.append(click)
        if self.gt_mask is not None:
            self.not_clicked_map[coords[0], coords[1]] = False

    def _remove_last_click(self):
        click = self.clicks_list.pop()
        coords = click.coords

        if click.is_positive:
            self.num_pos_clicks -= 1
        else:
            self.num_neg_clicks -= 1

        if self.gt_mask is not None:
            self.not_clicked_map[coords[0], coords[1]] = True

    def reset_clicks(self):
        if self.gt_mask is not None:
            self.not_clicked_map = np.ones_like(self.gt_mask, dtype=np.bool)

        self.num_pos_clicks = 0
        self.num_neg_clicks = 0

        self.clicks_list = []

    def get_state(self):
        return deepcopy(self.clicks_list)

    def set_state(self, state):
        self.reset_clicks()
        for click in state:
            self.add_click(click)

    def __len__(self):
        return len(self.clicks_list)


class Click:
    def __init__(self, is_positive, coords, indx=None):
        self.is_positive = is_positive
        self.coords = coords
        self.indx = indx

    @property
    def coords_and_indx(self):
        return (*self.coords, self.indx)

    def copy(self, **kwargs):
        self_copy = deepcopy(self)
        for k, v in kwargs.items():
            setattr(self_copy, k, v)
        return self_copy



import io
import logging
import time

import numpy as np
from torch.utils.tensorboard import SummaryWriter

# from datetime import datetime


# LOGGER_NAME = 'root'
# LOGGER_DATEFMT = '%Y-%m-%d %H:%M:%S'

# handler = logging.StreamHandler()

logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# logger.addHandler(handler)


# def add_logging(logs_path, prefix):
#     log_name = prefix + datetime.strftime(datetime.today(),
#       '%Y-%m-%d_%H-%M-%S') + '.log'
#     stdout_log_path = logs_path / log_name

#     fh = logging.FileHandler(str(stdout_log_path))
#     formatter = logging.Formatter(
#         fmt='(%(levelname)s) %(asctime)s: %(message)s',
#         datefmt=LOGGER_DATEFMT)
#     fh.setFormatter(formatter)
#     logger.addHandler(fh)


class TqdmToLogger(io.StringIO):
    logger = None
    level = None
    buf = ""

    def __init__(self, logger, level=None, mininterval=5):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO
        self.mininterval = mininterval
        self.last_time = 0

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        if (
            len(self.buf) > 0
            and time.time() - self.last_time > self.mininterval
        ):
            self.logger.log(self.level, self.buf)
            self.last_time = time.time()


class SummaryWriterAvg(SummaryWriter):
    def __init__(self, *args, dump_period=20, **kwargs):
        super().__init__(*args, **kwargs)
        self._dump_period = dump_period
        self._avg_scalars = dict()

    def add_scalar(self, tag, value, global_step=None, disable_avg=False):
        if disable_avg or isinstance(value, (tuple, list, dict)):
            super().add_scalar(tag, np.array(value), global_step=global_step)
        else:
            if tag not in self._avg_scalars:
                self._avg_scalars[tag] = ScalarAccumulator(self._dump_period)
            avg_scalar = self._avg_scalars[tag]
            avg_scalar.add(value)

            if avg_scalar.is_full():
                super().add_scalar(
                    tag, avg_scalar.value, global_step=global_step
                )
                avg_scalar.reset()


class ScalarAccumulator(object):
    def __init__(self, period):
        self.sum = 0
        self.cnt = 0
        self.period = period

    def add(self, value):
        self.sum += value
        self.cnt += 1

    @property
    def value(self):
        if self.cnt > 0:
            return self.sum / self.cnt
        else:
            return 0

    def reset(self):
        self.cnt = 0
        self.sum = 0

    def is_full(self):
        return self.cnt >= self.period

    def __len__(self):
        return self.cnt



from time import time

import numpy as np
import torch

from ritm_annotation.inference import utils
from ritm_annotation.inference.clicker import Clicker

if locals().get("get_ipython"):
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm


def evaluate_dataset(dataset, predictor, **kwargs):
    all_ious = []

    start_time = time()
    for index in tqdm(range(len(dataset)), leave=False):
        sample = dataset.get_sample(index)

        _, sample_ious, _ = evaluate_sample(
            sample.image, sample.gt_mask, predictor, sample_id=index, **kwargs
        )
        all_ious.append(sample_ious)
    end_time = time()
    elapsed_time = end_time - start_time

    return all_ious, elapsed_time


def evaluate_sample(
    image,
    gt_mask,
    predictor,
    max_iou_thr,
    pred_thr=0.49,
    min_clicks=1,
    max_clicks=20,
    sample_id=None,
    callback=None,
):
    clicker = Clicker(gt_mask=gt_mask)
    pred_mask = np.zeros_like(gt_mask)
    ious_list = []

    with torch.no_grad():
        predictor.set_input_image(image)

        for click_indx in range(max_clicks):
            clicker.make_next_click(pred_mask)
            pred_probs = predictor.get_prediction(clicker)
            pred_mask = pred_probs > pred_thr

            if callback is not None:
                callback(
                    image,
                    gt_mask,
                    pred_probs,
                    sample_id,
                    click_indx,
                    clicker.clicks_list,
                )

            iou = utils.get_iou(gt_mask, pred_mask)
            ious_list.append(iou)

            if iou >= max_iou_thr and click_indx + 1 >= min_clicks:
                break

        return (
            clicker.clicks_list,
            np.array(ious_list, dtype=np.float32),
            pred_probs,
        )



import numpy as np
import torch
import torch.nn as nn


class Initializer(object):
    def __init__(self, local_init=True, gamma=None):
        self.local_init = local_init
        self.gamma = gamma

    def __call__(self, m):
        if getattr(m, "__initialized", False):
            return

        if (
            isinstance(
                m,
                (
                    nn.BatchNorm1d,
                    nn.BatchNorm2d,
                    nn.BatchNorm3d,
                    nn.InstanceNorm1d,
                    nn.InstanceNorm2d,
                    nn.InstanceNorm3d,
                    nn.GroupNorm,
                    nn.SyncBatchNorm,
                ),
            )
            or "BatchNorm" in m.__class__.__name__
        ):
            if m.weight is not None:
                self._init_gamma(m.weight.data)
            if m.bias is not None:
                self._init_beta(m.bias.data)
        else:
            if getattr(m, "weight", None) is not None:
                self._init_weight(m.weight.data)
            if getattr(m, "bias", None) is not None:
                self._init_bias(m.bias.data)

        if self.local_init:
            object.__setattr__(m, "__initialized", True)

    def _init_weight(self, data):
        nn.init.uniform_(data, -0.07, 0.07)

    def _init_bias(self, data):
        nn.init.constant_(data, 0)

    def _init_gamma(self, data):
        if self.gamma is None:
            nn.init.constant_(data, 1.0)
        else:
            nn.init.normal_(data, 1.0, self.gamma)

    def _init_beta(self, data):
        nn.init.constant_(data, 0)


class Bilinear(Initializer):
    def __init__(self, scale, groups, in_channels, **kwargs):
        super().__init__(**kwargs)
        self.scale = scale
        self.groups = groups
        self.in_channels = in_channels

    def _init_weight(self, data):
        """Reset the weight and bias."""
        bilinear_kernel = self.get_bilinear_kernel(self.scale)
        weight = torch.zeros_like(data)
        for i in range(self.in_channels):
            if self.groups == 1:
                j = i
            else:
                j = 0
            weight[i, j] = bilinear_kernel
        data[:] = weight

    @staticmethod
    def get_bilinear_kernel(scale):
        """Generate a bilinear upsampling kernel."""
        kernel_size = 2 * scale - scale % 2
        scale = (kernel_size + 1) // 2
        center = scale - 0.5 * (1 + kernel_size % 2)

        og = np.ogrid[:kernel_size, :kernel_size]
        kernel = (1 - np.abs(og[0] - center) / scale) * (
            1 - np.abs(og[1] - center) / scale
        )

        return torch.tensor(kernel, dtype=torch.float32)


class XavierGluon(Initializer):
    def __init__(
        self, rnd_type="uniform", factor_type="avg", magnitude=3, **kwargs
    ):
        super().__init__(**kwargs)

        self.rnd_type = rnd_type
        self.factor_type = factor_type
        self.magnitude = float(magnitude)

    def _init_weight(self, arr):
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(arr)

        if self.factor_type == "avg":
            factor = (fan_in + fan_out) / 2.0
        elif self.factor_type == "in":
            factor = fan_in
        elif self.factor_type == "out":
            factor = fan_out
        else:
            raise ValueError("Incorrect factor type")
        scale = np.sqrt(self.magnitude / factor)

        if self.rnd_type == "uniform":
            nn.init.uniform_(arr, -scale, scale)
        elif self.rnd_type == "gaussian":
            nn.init.normal_(arr, 0, scale)
        else:
            raise ValueError("Unknown random type")



"""Python setup.py for ritm_annotation package"""
import io
import os
from setuptools import find_packages, setup
from distutils.core import Extension

try:
    import numpy as np
    use_numpy = True
except ImportError:
    use_numpy = False

try:
    from Cython.Build import cythonize
    use_cython = True
except ImportError:
    use_cython = False

def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("ritm_annotation", "VERSION")
    '0.1.0'
    >>> read("README.md")
    ...
    """

    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


def ext_modules():
    includes = []
    libraries = []

    if use_numpy:
        import numpy as np
        includes.append(np.get_include())
    if os.name == "posix":
        libraries.append("m")
    modules = []
    if use_cython:
        modules += cythonize(Extension(
            "*",
            ["ritm_annotation/**/*.pyx"],
            include_dirs=includes,
            libraries=libraries,
        ))
    return modules


setup(
    name="ritm_annotation",
    version=read("ritm_annotation", "VERSION"),
    description="Awesome ritm_annotation created by lucasew",
    url="https://github.com/lucasew/ritm_annotation/",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="lucasew",
    packages=find_packages(exclude=["tests", ".github"]),
    ext_modules=ext_modules(),
    install_requires=read_requirements("requirements.txt"),
    entry_points={
        "console_scripts": ["ritm_annotation = ritm_annotation.__main__:main"]
    },
    extras_require={"test": read_requirements("requirements-test.txt")},
    include_package_data=True
)



import numpy as np

cimport cython
cimport numpy as np
from libc.stdlib cimport free, malloc

ctypedef struct qnode:
    int row
    int col
    int layer
    int orig_row
    int orig_col

@cython.infer_types(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def get_dist_maps(np.ndarray[np.float32_t, ndim=2, mode="c"] points,
                  int height, int width, float norm_delimeter):
    cdef np.ndarray[np.float32_t, ndim=3, mode="c"] dist_maps = \
        np.full((2, height, width), 1e6, dtype=np.float32, order="C")

    cdef int *dxy = [-1, 0, 0, -1, 0, 1, 1, 0]
    cdef int i, j, x, y, dx, dy
    cdef qnode v
    cdef qnode *q = <qnode *> malloc((4 * height * width + 1) * sizeof(qnode))
    cdef int qhead = 0, qtail = -1
    cdef float ndist

    for i in range(points.shape[0]):
        x, y = round(points[i, 0]), round(points[i, 1])
        if x >= 0:
            qtail += 1
            q[qtail].row = x
            q[qtail].col = y
            q[qtail].orig_row = x
            q[qtail].orig_col = y
            if i >= points.shape[0] / 2:
                q[qtail].layer = 1
            else:
                q[qtail].layer = 0
            dist_maps[q[qtail].layer, x, y] = 0

    while qtail - qhead + 1 > 0:
        v = q[qhead]
        qhead += 1

        for k in range(4):
            x = v.row + dxy[2 * k]
            y = v.col + dxy[2 * k + 1]

            ndist = ((x - v.orig_row)/norm_delimeter) ** 2 + ((y - v.orig_col)/norm_delimeter) ** 2
            if (x >= 0 and y >= 0 and x < height and y < width and
                dist_maps[v.layer, x, y] > ndist):
                qtail += 1
                q[qtail].orig_col = v.orig_col
                q[qtail].orig_row = v.orig_row
                q[qtail].layer = v.layer
                q[qtail].row = x
                q[qtail].col = y
                dist_maps[v.layer, x, y] = ndist

    free(q)
    return dist_maps



import logging
import os
from pathlib import Path

import torch

from ritm_annotation.utils.exp import init_experiment
from ritm_annotation.utils.misc import load_module

logger = logging.getLogger(__name__)

COMMAND_DESCRIPTION = "Run from-scratch trains using off the shelf datasets"


def command(parser):
    parser.add_argument(
        "model_path", type=str, help="Path to the model script."
    )

    parser.add_argument(
        "-o",
        "--output",
        dest="experiment_path",
        type=Path,
        default=Path("."),
        help="Where to store experiment data",
    )

    parser.add_argument(
        "-n",
        "--num-epochs",
        dest="num_epochs",
        type=int,
        help="Amount of epochs",
    )

    parser.add_argument(
        "--exp-name",
        type=str,
        default="",
        help="Here you can specify the name of the experiment. "  # noqa:E501
        "It will be added as a suffix to the experiment folder.",
    )  # noqa:E501

    parser.add_argument(
        "--workers",
        type=int,
        default=4,
        metavar="N",
        help="Dataloader threads.",
    )

    parser.add_argument(
        "--batch-size",
        "-b",
        dest="batch_size",
        type=int,
        default=-1,
        help="You can override model batch size by specify positive number.",
    )  # noqa:E501

    parser.add_argument(
        "--ngpus",
        type=int,
        default=1,
        help="Number of GPUs. "
        'If you only specify "--gpus" argument, the ngpus value will be calculated automatically. '  # noqa:E501
        'You should use either this argument or "--gpus".',
    )  # noqa:E501

    parser.add_argument(
        "--gpus",
        type=str,
        default="",
        required=False,
        help='Ids of used GPUs. You should use either this argument or "--ngpus".',  # noqa: E501
    )  # noqa:E501

    parser.add_argument(
        "--resume-exp",
        type=str,
        default=None,
        help="The prefix of the name of the experiment to be continued. "  # noqa:E501
        'If you use this field, you must specify the "--resume-prefix" argument.',  # noqa:E501
    )  # noqa:E501

    parser.add_argument(
        "--resume-prefix",
        type=str,
        default="latest",
        help="The prefix of the name of the checkpoint to be loaded.",
    )  # noqa:E501

    parser.add_argument(
        "--start-epoch",
        type=int,
        default=0,
        help="The number of the starting epoch from which training will continue. "  # noqa:E501
        "(it is important for correct logging and learning rate)",
    )  # noqa:E501

    parser.add_argument(
        "--weights",
        type=str,
        default=None,
        help="Model weights will be loaded from the specified path if you use this argument.",  # noqa: E501
    )  # noqa:E501

    parser.add_argument(
        "--temp-model-path",
        type=str,
        default="",
        help="Do not use this argument (for internal purposes).",
    )  # noqa:E501

    parser.add_argument("--local_rank", type=int, default=0)

    def handle(args):
        model_path = Path(args.model_path)
        if args.temp_model_path != "":
            logger.debug(
                "Falling back to temp_model_path because model_path wasn't specified"  # noqa:E501
            )
            model_path = Path(args.temp_model_path)
        if not model_path.is_absolute():
            model_path = (
                Path(__file__).parent.parent.parent / "models" / model_path
            )
        model_path = model_path.resolve()
        args.model_path = model_path
        logger.debug(f"Final model path: '{model_path}'")

        model_script = load_module(model_path)

        model_base_name = model_script.__dict__.get("MODEL_NAME")

        args.distributed = "WORLD_SIZE" in os.environ
        cfg = init_experiment(args, model_base_name)
        cfg = load_cfg_from_env(cfg, os.environ)

        torch.backends.cudnn.benchmark = True
        torch.multiprocessing.set_sharing_strategy("file_system")
        logger.debug("Basic validations passed")
        model, model_cfg = model_script.init_model(cfg)
        trainer = model_script.get_trainer(model, cfg, model_cfg)
        if args.num_epochs is None:
            args.num_epochs = model_cfg.default_num_epochs
        trainer.run(num_epochs=args.num_epochs)

    return handle



from .zoom_in import ZoomIn, get_roi_image_nd


class LimitLongestSide(ZoomIn):
    def __init__(self, max_size=800):
        super().__init__(target_size=max_size, skip_clicks=0)

    def transform(self, image_nd, clicks_lists):
        assert image_nd.shape[0] == 1 and len(clicks_lists) == 1
        image_max_size = max(image_nd.shape[2:4])
        self.image_changed = False

        if image_max_size <= self.target_size:
            return image_nd, clicks_lists
        self._input_image = image_nd

        self._object_roi = (0, image_nd.shape[2] - 1, 0, image_nd.shape[3] - 1)
        self._roi_image = get_roi_image_nd(
            image_nd, self._object_roi, self.target_size
        )
        self.image_changed = True

        tclicks_lists = [self._transform_clicks(clicks_lists[0])]
        return self._roi_image, tclicks_lists
